# =============================================================================
# Minet Apply Scraper Function
# =============================================================================
#
# Function taking a scraper definition and applying its logic recursively
# to yield its result.
#
from typing import Optional

import textwrap
import soupsieve
from bs4 import Tag
from ebbe import getpath
from functools import partial

from minet.scrape.std import get_default_evaluation_context, get_display_text
from minet.scrape.constants import EXTRACTOR_NAMES
from minet.scrape.utils import get_sel, get_iterator
from minet.scrape.exceptions import (
    ScraperEvalError,
    ScraperEvalTypeError,
    ScraperEvalNoneError,
    NotATableError,
)

DEFAULT_CONTEXT = {}
DATA_TYPES = (str, int, float, bool, list, dict)

nested_getter = partial(getpath, split_char=".", parse_indices=True, attributes=True)


def is_list_of_tags(value):
    if not isinstance(value, list):
        return False

    return all(isinstance(tag, Tag) for tag in value)


def is_valid_iterator_eval_output(value):
    return isinstance(value, str) or is_list_of_tags(value)


def merge_contexts(global_context, local_context):
    if global_context is None:
        return local_context

    for k, v in global_context.items():
        if k not in local_context:
            local_context[k] = v

    return local_context


def extract(element, extractor_name):
    if extractor_name == "text":
        return element.get_text().strip()

    if extractor_name == "display_text":
        return get_display_text(element)

    if extractor_name == "html" or extractor_name == "inner_html":
        return element.decode_contents().strip()

    if extractor_name == "outer_html":
        return str(element).strip()

    raise TypeError('Unknown "%s" extractor' % extractor_name)


class EvaluationScope(object):
    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __getattr__(self, key):
        return self.__dict__.get(key)


EVAL_CONTEXT = get_default_evaluation_context()


# NOTE: this is not threadsafe, but it does not have to be
def eval_expression(
    expression,
    element=None,
    elements=None,
    value=None,
    context=None,
    root=None,
    scope=None,
    path=None,
    expect=None,
    check=None,
    allow_none=False,
):
    if callable(expression):
        try:
            result = expression(
                element=element,
                elements=elements,
                value=value,
                context=context,
                root=root,
            )
        except Exception as e:
            raise ScraperEvalError(reason=e, path=path, expression=expression)
    else:
        # Local variables
        EVAL_CONTEXT["element"] = element
        EVAL_CONTEXT["elements"] = elements
        EVAL_CONTEXT["value"] = value

        # Context
        EVAL_CONTEXT["context"] = context
        EVAL_CONTEXT["root"] = root
        EVAL_CONTEXT["scope"] = scope

        if "\n" in expression:
            # Multiline expression
            scope = {}

            wrapped_expression = (
                "def __run__():\n%s\n__return_value__ = __run__()"
                % textwrap.indent(expression, "  ")
            )

            try:
                exec(wrapped_expression, EVAL_CONTEXT, scope)
                result = scope["__return_value__"]
            except Exception as e:
                raise ScraperEvalError(reason=e, path=path, expression=expression)
        else:
            # Simple expression
            try:
                result = eval(expression, EVAL_CONTEXT, None)
            except Exception as e:
                raise ScraperEvalError(reason=e, path=path, expression=expression)

    if not allow_none and result is None:
        raise ScraperEvalNoneError(path=path, expression=expression)

    if (expect is not None or check is not None) and result is not None:
        if not (check(result) if check else isinstance(result, expect)):
            raise ScraperEvalTypeError(
                path=path, expression=expression, expected=expect, got=result
            )

    return result


def tabulate(element, headers_inference: Optional[str] = "th", headers=None, path=None):
    if element.name != "table":
        raise NotATableError(path=path)

    def generator():
        nonlocal headers

        body = element.find("tbody", recursive=False)
        head = element.find("thead", recursive=False)

        if body is None:
            body = element

        if head is None:
            head = element

        trs = body.select("tr:has(td)", recursive=False)
        ths = head.select("tr > th", recursive=False)

        if headers is None:
            if headers_inference == "th":
                headers = [th.get_text() for th in ths]

        if headers is not None:
            for tr in trs:
                yield {
                    headers[i]: td.get_text()
                    for i, td in enumerate(tr.find_all("td", recursive=False))
                }
        else:
            for tr in trs:
                yield [td.get_text() for td in tr.find_all("td", recursive=False)]

    return generator()


def interpret_scraper(scraper, element, root=None, context=None, path=[], scope=None):
    if scope is None:
        scope = EvaluationScope()

    # Is this a tail call of item?
    if isinstance(scraper, str):
        if scraper in EXTRACTOR_NAMES:
            return extract(element, scraper)

        return element.get(scraper)

    sel = get_sel(scraper)
    iterator = get_iterator(scraper)

    # First we need to solve local selection
    if sel is not None:
        element = soupsieve.select_one(sel, element)
    elif "sel_eval" in scraper:
        evaluated_sel = eval_expression(
            scraper["sel_eval"],
            element=element,
            elements=[],
            context=context,
            root=root,
            path=path + ["sel_eval"],
            expect=(Tag, str),
            allow_none=True,
            scope=scope,
        )

        if isinstance(evaluated_sel, str):
            element = soupsieve.select_one(evaluated_sel, element)
        else:
            element = evaluated_sel

    if element is None:
        return None

    # Then we need to solve iterator
    single_value = True

    if iterator is not None:
        single_value = False
        elements = soupsieve.select(iterator, element)
    elif "iterator_eval" in scraper:
        single_value = False
        evaluated_elements = eval_expression(
            scraper["iterator_eval"],
            element=element,
            elements=[],
            context=context,
            root=root,
            path=path + ["iterator_eval"],
            check=is_valid_iterator_eval_output,
            scope=scope,
        )

        if isinstance(evaluated_elements, str):
            elements = soupsieve.select(evaluated_elements, element)
        else:
            elements = evaluated_elements
    else:
        elements = [element]

    # Handling local context
    if "set_context" in scraper:
        local_context = {}

        for k, field_scraper in scraper["set_context"].items():
            local_context[k] = interpret_scraper(
                field_scraper,
                element,
                root=root,
                context=context,
                path=path + ["set_context", k],
                scope=scope,
            )

        context = merge_contexts(context, local_context)

    # Actual iteration
    acc = None if single_value else []

    already_seen = set() if "uniq" in scraper and not single_value else None

    for element in elements:
        value = None

        # Do we have fields?
        if "fields" in scraper:
            value = {}

            for k, field_scraper in scraper["fields"].items():
                value[k] = interpret_scraper(
                    field_scraper,
                    element,
                    root=root,
                    context=context,
                    path=path + ["fields", k],
                    scope=scope,
                )

        # Do we have a scalar?
        elif "item" in scraper:
            # Default value is text
            value = interpret_scraper(
                scraper["item"],
                element,
                root=root,
                context=context,
                path=path + ["item"],
                scope=scope,
            )

        else:
            if "attr" in scraper:
                value = element.get(scraper["attr"])
            elif "extract" in scraper:
                value = extract(element, scraper["extract"])
            elif "get_context" in scraper:
                value = nested_getter(context, scraper["get_context"])
            elif "default" not in scraper:
                # Default value is text
                value = extract(element, "text")

            # Eval?
            if "eval" in scraper:
                value = eval_expression(
                    scraper["eval"],
                    element=element,
                    elements=elements,
                    value=value,
                    context=context,
                    root=root,
                    path=path + ["eval"],
                    expect=DATA_TYPES,
                    allow_none=True,
                    scope=scope,
                )

        # Default value after all?
        if "default" in scraper and value is None:
            value = scraper["default"]

        if single_value:
            acc = value
        else:
            # Filtering?
            if "filter_eval" in scraper:
                passed_filter = eval_expression(
                    scraper["filter_eval"],
                    element=element,
                    elements=elements,
                    value=value,
                    context=context,
                    root=root,
                    path=path + ["filter_eval"],
                    expect=bool,
                    allow_none=True,
                    scope=scope,
                )

                if not passed_filter:
                    continue

            if "filter" in scraper:
                filtering_clause = scraper["filter"]

                if filtering_clause is True and not value:
                    continue

                if isinstance(filtering_clause, str) and not nested_getter(
                    value, filtering_clause
                ):
                    continue

            if "uniq" in scraper:
                assert already_seen is not None

                uniq_clause = scraper["uniq"]
                k = value

                if uniq_clause is True and value in already_seen:
                    continue

                if isinstance(uniq_clause, str):
                    k = nested_getter(value, uniq_clause)

                    if k in already_seen:
                        continue

                already_seen.add(k)

            acc.append(value)

    return acc



from typing import TYPE_CHECKING

from minet.exceptions import MinetError

if TYPE_CHECKING:
    from minet.crawl.types import CrawlJob
    from minet.web import Response


class CrawlerError(MinetError):
    pass


class CrawlerAlreadyFinishedError(CrawlerError):
    pass


class CrawlerSpiderProcessError(CrawlerError):
    def __init__(self, reason: Exception, job: "CrawlJob", response: "Response"):
        self.reason = reason
        self.job = job
        self.response = response
        super().__init__(str(reason))



from typing import (
    Dict,
    List,
    Tuple,
    Iterable,
    Iterator,
    Optional,
    Callable,
    Union,
    Literal,
)

import sqlite3
import pickle
from os import makedirs
from os.path import join, isfile
from shutil import rmtree
from queue import Empty
from threading import Lock, Condition
from contextlib import contextmanager
from datetime import datetime
from random import randint
from dataclasses import dataclass
from quenouille.constants import TIMER_EPSILON

from minet.crawl.types import CrawlJob
from minet.utils import iterate_over_sqlite_cursor

AnyThrottle = Union[float, Callable[[CrawlJob], float]]
AnyParallelism = Union[int, Callable[[CrawlJob], int]]


def now() -> float:
    return datetime.now().timestamp()


# NOTE: synchronous=normal is probably alright for our use-case
SQL_PRAGMAS = """
PRAGMA journal_mode=wal;
PRAGMA synchronous=normal;
"""

# NOTE about status:
# status=0 is to do
# status=1 is doing
# status=2 is done, but kept until cleanup

# NOTE: the multi-index on ("status", "priority", "index") seems
# to be the better one according to the query planner
# NOTE: we are not relying on AUTOINCREMENT because it seems to
# be bad for performance and because our queue can handle it on
# its own. This means that the job having the max index must
# never be deleted for this queue to be able to resume without issue.
SQL_CREATE = """
CREATE TABLE "queue" (
    "index" INTEGER PRIMARY KEY,
    "status" INTEGER NOT NULL DEFAULT 0,
    "id" TEXT NOT NULL,
    "url" TEXT NOT NULL,
    "group" TEXT,
    "depth" INTEGER NOT NULL,
    "spider" TEXT,
    "priority" INTEGER NOT NULL,
    "data" BLOB,
    "parent" TEXT
);
CREATE INDEX "idx_queue_multi" ON "queue" ("status", "priority", "index");

CREATE TABLE "throttle" (
    "group" TEXT PRIMARY KEY,
    "timestamp" REAL NOT NULL
) WITHOUT ROWID;
CREATE INDEX "idx_throttle_timestamp" ON "throttle" ("timestamp");

CREATE TABLE "parallelism" (
    "group" TEXT PRIMARY KEY,
    "count" INTEGER NOT NULL,
    "allowed" INTEGER NOT NULL
) WITHOUT ROWID;
"""

SQL_INSERT_JOB = """
INSERT INTO "queue" (
    "index",
    "id",
    "url",
    "group",
    "depth",
    "spider",
    "priority",
    "data",
    "parent"
) VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?);
"""

# NOTE: we thought about the opportunity of using
# RIGHT JOIN as further optimization but it cannot work
# without keeping a reference to all the groups in the "throttle"
# and "parallelism" table at all time and this is hardly optimal
# regarding disk space and query time, in spite of indices.
SQL_GET_JOB = """
SELECT
    "queue"."index",
    "queue"."id",
    "queue"."url",
    "queue"."group",
    "queue"."depth",
    "queue"."spider",
    "queue"."priority",
    "queue"."data",
    "queue"."parent"
FROM "queue"
LEFT JOIN "throttle" ON "queue"."group" = "throttle"."group"
LEFT JOIN "parallelism" ON "queue"."group" = "parallelism"."group"
WHERE (
   "queue"."status" = 0
   AND ("throttle"."timestamp" IS NULL OR "throttle"."timestamp" <= ?)
   AND ("parallelism"."count" IS NULL OR "parallelism"."count" < "parallelism"."allowed")
)
ORDER BY "priority" ASC, "index" %s
LIMIT 1;
"""

SQL_DUMP = """
SELECT
    "index",
    "id",
    "url",
    "group",
    "depth",
    "spider",
    "priority",
    "data",
    "parent",
    "status"
FROM "queue"
ORDER BY "index";
"""

# NOTE: we thought about having a single allowance column that would
# be decremented when worked by threads but it means we need to keep
# a reference for each group in the "parallelism" table at all time
# for this to work and this is hardly optimal regarding disk space
# and query time, in spite of indices.
SQL_INCREMENT_PARALLELISM = """
INSERT OR REPLACE INTO "parallelism" ("group", "count", "allowed") VALUES (
    ?,
    COALESCE((SELECT ("count" + 1) FROM "parallelism" WHERE "group" = ? LIMIT 1), 1),
    ?
)
"""

SQL_UPDATE_THROTTLE = """
INSERT OR REPLACE INTO "throttle" ("group", "timestamp") VALUES (?, ?);
"""

CrawlerQueueJobStatus = Literal["todo", "doing", "done"]


@dataclass
class CrawlerQueueRecord:
    index: int
    status: CrawlerQueueJobStatus
    job: CrawlJob

    @staticmethod
    def fieldnames():
        return ["index", "status"] + CrawlJob.fieldnames()

    def __csv_row__(self) -> List:
        row = [self.index, self.status] + list(self.job.__csv_row__())
        return row


# NOTE: this exception is only used when we need to manually
# unblock the queue, e.g. when quenouille must teardown
class BrokenCrawlerQueue(Exception):
    pass


# TODO: tests with null group


# NOTE: this queue can be used by `quenouille` but since it handles
# group parallelism and throttling on its own, we must shunt it on
# `quenouille` side. Also, keep in mind that buffer_size must be
# set to 0 to keep the queue ordering consistent and bypass `quenouille`
# optimistic buffer that is tailored for lazy iterable, not processing
# queues.
# NOTE: this queue does not follow python's queue.Queue semantics exactly
# because its #.get method can block, but will raise Empty when drained.
# This is fine with `quenouille` if we don't forget to call #.unblock
# on panic.
# NOTE: sqlite needs to serialize writes, which means it is not particularly
# useful in our case to allow for concurrent access to the queue because most
# of its operations need to update the database somehow. This also means we
# can rely on a single lock to make multithreaded transactions safe.
class CrawlerQueue:
    # Params
    persistent: bool
    resuming: bool
    is_lifo: bool
    group_parallelism: AnyParallelism
    throttle: AnyThrottle

    # State
    tasks: Dict[str, int]
    connection: sqlite3.Connection
    counter: int
    cleanup_interval: int
    vacuum_interval: int
    waiter: Condition
    currently_waiting_count: int
    transaction_lock: Lock
    broken: bool

    def __init__(
        self,
        path: Optional[str] = None,
        db_name: str = "queue.db",
        resume: bool = False,
        inspect: bool = False,
        lifo: bool = False,
        group_parallelism: AnyParallelism = 1,
        throttle: AnyThrottle = 0,
        cleanup_interval: int = 1000,
        vacuum_interval: int = 10_000,
    ):
        self.persistent = True
        self.resuming = False

        if path is None:
            if inspect:
                raise TypeError("cannot inspect a memory queue")

            self.persistent = False
            # NOTE: I am not sure this is advisable most of the time
            full_path = "file:%i-%i.db?mode=memory&cache=shared" % (
                id(self),
                randint(0, 2**32 - 1),
            )
        else:
            full_path = join(path, db_name)

            if inspect:
                if not isfile(full_path):
                    raise RuntimeError(
                        "sqlite database for queue does not exist and cannot be inspected"
                    )
            elif not resume:
                rmtree(path, ignore_errors=True)
            elif isfile(full_path):
                self.resuming = True

            makedirs(path, exist_ok=True)

        self.tasks = {}

        self.is_lifo = lifo

        self.group_parallelism = group_parallelism
        self.throttle = throttle

        self.cleanup_interval = cleanup_interval
        self.vacuum_interval = vacuum_interval
        self.counter = 0
        self.current_task_done_count = 0

        self.waiter = Condition()
        self.currently_waiting_count = 0
        self.transaction_lock = Lock()
        self.broken = False

        self.connection = sqlite3.connect(full_path, check_same_thread=False)

        # NOTE: it's seems it is safer and common practice to
        # reexecute pragmas each time because they might not
        # be stored persistently in some instances.
        self.connection.executescript(SQL_PRAGMAS)
        self.connection.commit()

        if inspect:
            return

        # Setup
        with self.transaction() as cursor:
            if not self.resuming:
                cursor.executescript(SQL_CREATE)
            else:
                # We need to restart our counter
                cursor.execute('SELECT max("index") FROM "queue";')
                self.counter = cursor.fetchone()[0] + 1

                # We need to clear status=1
                # NOTE: we don't clear status > 1 just in case we need to use those
                # to indicate something else
                cursor.execute('UPDATE "queue" SET "status" = 0 WHERE "status" = 1;')

                # We can safely drop parallelism info as it is bound to runtime
                cursor.execute('DELETE FROM "parallelism";')

                # Cleanup throttle a bit
                cursor.execute('DELETE FROM "throttle" WHERE "timestamp" < ?', (now(),))

                cursor.connection.commit()
                cursor.execute("VACUUM;")

    @contextmanager
    def transaction(self):
        cursor = None
        try:
            with self.transaction_lock, self.connection:
                cursor = self.connection.cursor()
                yield cursor
        finally:
            if cursor is not None:
                cursor.close()

    def explain_query_plan(self, sql: str) -> str:
        if sql == "get":
            sql = SQL_GET_JOB % ("ASC" if not self.is_lifo else "DESC")

        sql = sql.replace("?", "1")

        with self.transaction() as cursor:
            cursor.execute("EXPLAIN QUERY PLAN %s" % sql)

            return "\n".join(row[3] for row in iterate_over_sqlite_cursor(cursor))

    def __count(self, cursor: sqlite3.Cursor) -> int:
        cursor.execute('SELECT count(*) FROM "queue" WHERE "status" = 0;')
        return cursor.fetchone()[0]

    def qsize(self) -> int:
        with self.transaction() as cursor:
            return self.__count(cursor)

    def __len__(self) -> int:
        return self.qsize()

    def put_many(self, jobs: Iterable[CrawlJob]) -> int:
        with self.transaction() as cursor:
            rows = []

            for job in jobs:
                rows.append(
                    (
                        self.counter,
                        job.id,
                        job.url,
                        job.group,
                        job.depth,
                        job.spider,
                        job.priority,
                        pickle.dumps(job.data) if job.data is not None else None,
                        job.parent,
                    )
                )
                self.counter += 1

            # NOTE: executemany prepares the statement for us
            cursor.executemany(SQL_INSERT_JOB, rows)
            count = cursor.rowcount

        # NOTE: we notify the waiter because adding jobs to the queue means
        # there might be one we can do right now
        with self.waiter:
            self.waiter.notify()

        return count

    def put(self, job: CrawlJob) -> None:
        self.put_many((job,))

    # NOTE: we will need to cheat a little bit to work with quenouille here.
    # This method will actually block but raise Empty if the queue is drained.
    # We will also need to handle throttling and group parallelism
    # on our own and use buffer_size=0 on quenouille's size to bypass the
    # optimistic buffer that trumps the true ordering of the given queue.
    def get(self, block=True) -> CrawlJob:
        if block:
            raise NotImplementedError

        need_to_wait = False
        need_to_wait_for_at_least = None

        while True:
            # Waiting?
            # NOTE: we wait here so we can: 1. avoid recursion and 2. release
            # the transaction lock to avoid an obvious deadlock.
            if need_to_wait:
                with self.waiter:
                    self.currently_waiting_count += 1
                    self.waiter.wait(need_to_wait_for_at_least)
                    self.currently_waiting_count -= 1

                need_to_wait = False
                need_to_wait_for_at_least = None

                # NOTE: this only happens when unblocking manually and
                # usually the queue won't be used anymore afterwards.
                # If we finally understand the contrary, we should
                # add some #.repair method or change threading paradigm.
                if self.broken:
                    raise BrokenCrawlerQueue

            with self.transaction() as cursor:
                cursor.execute(
                    SQL_GET_JOB % ("ASC" if not self.is_lifo else "DESC"),
                    (now(),),
                )
                row = cursor.fetchone()

                if row is None:
                    # Queue really is drained
                    if self.__count(cursor) == 0:
                        raise Empty

                    # We may need to wait for a suitable job
                    # NOTE: here we may wait either for one slot to become
                    # open wrt parallelism or enough time wrt throttling
                    need_to_wait = True

                    cursor.execute('SELECT min("timestamp") FROM "throttle" LIMIT 1;')
                    throttle_row = cursor.fetchone()

                    if throttle_row is not None and throttle_row[0] is not None:
                        need_to_wait_for_at_least = (
                            max(0, throttle_row[0] - now()) + TIMER_EPSILON
                        )

                    continue

                index = row[0]

                # NOTE: sqlite does not always support LIMIT on UPDATE
                cursor.execute(
                    'UPDATE "queue" SET "status" = 1 WHERE "index" = ?;',
                    (index,),
                )

                job = CrawlJob(
                    row[2],
                    id=row[1],
                    group=row[3],
                    depth=row[4],
                    spider=row[5],
                    priority=row[6],
                    data=pickle.loads(row[7]) if row[7] is not None else None,
                    parent=row[8],
                )

                if job.group is not None:
                    allowed = self.group_parallelism

                    if callable(allowed):
                        allowed = allowed(job)

                    cursor.execute(
                        SQL_INCREMENT_PARALLELISM,
                        (job.group, job.group, allowed),
                    )

                self.tasks[job.id] = index

                return job

    def get_nowait(self) -> CrawlJob:
        return self.get(False)

    def unblock(self) -> None:
        with self.waiter:
            if self.currently_waiting_count == 0:
                return

            self.broken = True
            self.waiter.notify_all()

    def worked_groups(self) -> Dict[str, Tuple[int, int]]:
        g = {}

        with self.transaction() as cursor:
            cursor.execute(
                'SELECT "group", "count", "allowed" FROM "parallelism" WHERE "count" > 0;'
            )

            for row in iterate_over_sqlite_cursor(cursor):
                g[row[0]] = (row[1], row[2])

        return g

    def dump(self) -> Iterator[CrawlerQueueRecord]:
        with self.transaction() as cursor:
            cursor.execute(SQL_DUMP)

            for row in iterate_over_sqlite_cursor(cursor):
                job = CrawlJob(
                    row[2],
                    id=row[1],
                    group=row[3],
                    depth=row[4],
                    spider=row[5],
                    priority=row[6],
                    data=pickle.loads(row[7]) if row[7] is not None else None,
                    parent=row[8],
                )

                status_code = row[9]
                status = "todo"

                if status_code == 1:
                    status = "doing"
                elif status_code == 2:
                    status = "done"

                yield CrawlerQueueRecord(index=row[0], status=status, job=job)

    def __cleanup(self, cursor: sqlite3.Cursor) -> None:
        # NOTE: we make sure to keep the last index given to ensure we can resume safely
        cursor.execute(
            'DELETE FROM "queue" WHERE "status" = 2 AND "index" <> (SELECT max("index") FROM "queue" ORDER BY "index");'
        )
        cursor.execute('DELETE FROM "parallelism" WHERE "count" < 1;')
        cursor.execute('DELETE FROM "throttle" WHERE "timestamp" < ?', (now(),))

    def __clear(self, cursor: sqlite3.Cursor) -> None:
        cursor.execute('DELETE FROM "queue";')
        cursor.execute('DELETE FROM "parallelism";')
        cursor.execute('DELETE FROM "throttle";')

    def __vacuum_and_analyze(self, cursor: sqlite3.Cursor) -> None:
        cursor.execute("PRAGMA analysis_limit=1000;")
        cursor.execute("PRAGMA optimize;")
        cursor.connection.commit()
        cursor.execute("VACUUM;")
        cursor.connection.commit()

    def cleanup(self) -> None:
        with self.transaction() as cursor:
            self.__cleanup(cursor)

    def clear(self) -> None:
        with self.transaction() as cursor:
            self.__clear(cursor)

    # NOTE: there is a subtle difference between #.release_group
    # and #.task_done. The first indicates to the queue that we
    # finished network actions related to the given group so
    # we can update parallelism & throttling info. The latter that
    # the task is completely done, i.e. after outputs were flushed
    # in the crawler loop so that we can safely forget about the given
    # task. This is important to ensure atomicity as well as possible.
    # Without creating backpressure on the queue's constraints.
    def release_group(self, job: CrawlJob) -> None:
        with self.transaction() as cursor:
            if job.id not in self.tasks:
                raise RuntimeError("job is not being worked")

            cursor.execute(
                'UPDATE "parallelism" SET "count" = "count" - 1 WHERE "group" = ?;',
                (job.group,),
            )

            # NOTE: null group is not throttled, the same as `quenouille`
            if job.group is not None:
                throttle = self.throttle

                if callable(throttle):
                    throttle = throttle(job)

                if throttle > 0:
                    cursor.execute(SQL_UPDATE_THROTTLE, (job.group, now() + throttle))

        # We notify one waiter that parallelism was updated
        with self.waiter:
            self.waiter.notify()

    @contextmanager
    def group_releaser(self, job: CrawlJob):
        try:
            yield
        finally:
            self.release_group(job)

    def task_done(self, job: CrawlJob) -> None:
        with self.transaction() as cursor:
            index = self.tasks.pop(job.id, None)

            if index is None:
                raise RuntimeError("job is not being worked")

            cursor.execute(
                'UPDATE "queue" SET "status" = 2 WHERE "index" = ?;', (index,)
            )

            self.current_task_done_count += 1

            if self.current_task_done_count % self.cleanup_interval == 0:
                self.__cleanup(cursor)

            if self.current_task_done_count % self.vacuum_interval == 0:
                self.__vacuum_and_analyze(cursor)

    def close(self) -> None:
        self.connection.close()

    def __del__(self) -> None:
        self.close()



# =============================================================================
# Minet Scrape Module
# =============================================================================
#
# Module exposing utilities related to minet's scraping DSL.
#
from minet.scrape.classes.definition import scrape, DefinitionScraper, validate
from minet.scrape.soup import WonderfulSoup, SelectionError, ExtractionError
from minet.scrape.regex import (
    extract_encodings_from_xml,
    extract_canonical_link,
    extract_javascript_relocation,
    extract_meta_refresh,
)

__all__ = [
    "scrape",
    "DefinitionScraper",
    "validate",
    "WonderfulSoup",
    "SelectionError",
    "ExtractionError",
    "extract_encodings_from_xml",
    "extract_canonical_link",
    "extract_javascript_relocation",
    "extract_meta_refresh",
]



# =============================================================================
# Minet Google CLI Action
# =============================================================================
#
# Logic of the `google` action.
#
from minet.cli.argparse import command

GOOGLE_SHEETS_SUBCOMMAND = command(
    "sheets",
    "minet.cli.google.sheets",
    title="Minet Google Sheets Command",
    description="""
        Grab the given google spreadsheet as a CSV file from
        its url, its sharing url or id.

        It can access public spreadsheets without issues, but to
        you will need to tell the command how to retrieve Google
        drive authentication cookies to also be able to access
        private ones.

        Also note that by default, the command will try to access
        your spreadsheet using your first 4 connected Google
        accounts.

        If you have more connected accounts or know beforehand
        to which account some spreadsheets are tied to, be sure
        to give --authuser.
    """,
    epilog="""
        Examples:

        . Exporting from the spreadsheet id:
            $ minet google sheets 1QXQ1yaNYrVUlMt6LQ4jrLGt_PvZI9goozYiBTgaC4RI > file.csv

        . Using your Firefox authentication cookies:
            $ minet google sheets --cookie firefox 1QXQ1yaNYrVUlMt6LQ4jrLGt_PvZI9goozYiBTgaC4RI > file.csv
    """,
    arguments=[
        {
            "name": "url",
            "help": "Url, sharing url or id of the spreadsheet to export.",
        },
        {
            "flags": ["-a", "--authuser"],
            "help": "Connected google account number to use.",
            "type": int,
        },
        {
            "flags": ["-c", "--cookie"],
            "help": 'Google Drive cookie or browser from which to extract it (supports "firefox", "chrome", "chromium", "opera" and "edge").',
        },
    ],
)

GOOGLE_COMMAND = command(
    "google",
    "minet.cli.google",
    "Minet Google Command",
    description="""
        Collect data from Google.
    """,
    subcommands=[GOOGLE_SHEETS_SUBCOMMAND],
)



from minet.cli.argparse import command, InputAction

DUMP_QUEUE_COMMAND = command(
    "dump-queue",
    "minet.cli.dump.dump_queue",
    title="Minet Dump Queue Command",
    description="""
        Debug command that can be used to dump a crawler's queue
        and see what's inside.
    """,
    epilog="""
        Examples:

        . Dumping a queue:
            $ minet dump-queue ./path-to-queue-directory
    """,
    arguments=[
        {"name": "queue_dir", "help": "Path to the queue's directory."},
    ],
)

DUMP_URLS_COMMAND = command(
    "dump-urls",
    "minet.cli.dump.dump_urls",
    title="Minet Dump Urls Command",
    description="""
        Debug command that can be used to dump a crawler's url store
        and see what's inside.
    """,
    epilog="""
        Examples:

        . Dumping a url store:
            $ minet dump-urls ./path-to-urls-directory
    """,
    arguments=[
        {"name": "urls_dir", "help": "Path to the url store's directory."},
    ],
)

DUMP_GRAPH_COMMAND = command(
    "dump-graph",
    "minet.cli.dump.dump_graph",
    title="Minet Dump Graph Command",
    description="""
        Debug command that can be used to dump a crawler's graph from
        a jobs csv and manipulate it as a network.

        The export file is in json with a Graphology format.
    """,
    epilog="""
        Examples:

        . Dumping a graph file from crawler jobs:
            $ minet dump-graph -i ./jobs.csv > graph.json
    """,
    select=True,
    arguments=[
        {
            "flags": ["-i", "--input"],
            "help": "Path to crawler's jobs csv file.",
            "action": InputAction,
        },
        {
            "flags": ["-l", "--label"],
            "help": "Specify the CSV column to use for graph's nodes label",
            "default": "url",
        },
    ],
)



from typing import List, Any

from minet.crawl import CrawlResult
from minet.crawl.focus import FocusCrawlInfo, FocusSpider

from minet.cli.crawl.crawl import crawl_action
from minet.cli.loading_bar import LoadingBar


def action(cli_args):
    additional_fieldnames = FocusCrawlInfo.CRAWL_RESULT_ADDENDUM_FIELDNAMES
    padding = [None] * len(additional_fieldnames)

    spider = FocusSpider(
        regex_content=cli_args.content_filter,
        invert_content_match=cli_args.invert_content_match,
        regex_url=cli_args.url_filter,
        invert_url_match=cli_args.invert_url_match,
        irrelevant_continue=cli_args.irrelevant_continue,
        only_html=cli_args.only_html,
        extract=cli_args.extract,
        extraction_fields=cli_args.extraction_fields,
    )

    def format_job_row_addendum(result: CrawlResult[Any, FocusCrawlInfo]) -> List:
        info = result.data

        if not info:
            return padding

        return info.as_crawl_result_csv_addendum()

    def result_callback(
        cli_args, loading_bar: LoadingBar, result: CrawlResult[Any, FocusCrawlInfo]
    ):
        if result.data is not None and result.data.relevant:
            loading_bar.inc_stat("relevant", style="success")
        else:
            loading_bar.inc_stat("irrelevant", style="warning")

    return crawl_action(
        cli_args,
        target=spider,
        additional_job_fieldnames=additional_fieldnames,
        format_job_row_addendum=format_job_row_addendum,
        result_callback=result_callback,
    )



from casanova import IndexedResumer

from minet.cli.argparse import command


def resolve_arguments(cli_args):
    if cli_args.column is None:
        cli_args.column = "path"


EXTRACT_COMMAND = command(
    "extract",
    "minet.cli.extract.extract",
    title="Minet Extract Command",
    description="""
        Use multiple processes to extract the main textual content and various
        metadata item from large batches of HTML files easily and efficiently.

        Extraction of main content is performed using the `trafilatura` library
        by Adrien Barbaresi.

        Please note that this kind of main content extraction is heavily geared
        towards press articles and might not be suited to other kinds of web pages.

        More information about the library can be found here:
        https://github.com/adbar/trafilatura

        Note that this command has been geared towards working in tandem with
        the fetch command. This means the command expects, by default, CSV files
        containing columns like "path", "http_status", "encoding" etc. that
        you can find in a fetch command CSV report.

        This said, you can of course feed this command any kind of CSV data,
        and use dedicated flags such as --status-column, --body-column to
        to inform the command about your specific table.

        The command is also able to work on glob patterns, such as: "downloaded/**/*.html",
        and can also be fed CSV columns containing HTML content directly if
        required.
    """,
    epilog="""

        Columns being added to the output:

        . "extract_original_index": index of the line in the original file (the output will be
            arbitrarily ordered since multiple requests are performed concurrently).
        . "extract_error": any error that happened when extracting content.
        . "canonical_url": canonical url of target html, extracted from
            link[rel=canonical].
        . "title": title of the web page, from <title> usually.
        . "description": description of the web page, as found in its
            metadata.
        . "content": main content of the web page as extracted.
        . "comments": comment text whenever the heuristics succeeds in
            identifying them.
        . "author": inferred author of the web page article when found in
            its metadata.
        . "categories": list of categories extracted from the web page's
            metadata, separated by "|".
        . "tags": list of tags extracted from the web page's metadata,
            separated by "|".
        . "date": date of publication of the web page article when found in
            its metadata.
        . "sitename": canonical name as declared by the website.

        Examples:

        . Extracting content from a single file on disk:
            $ minet extract ./path/to/file.html

        . Extracting content from a `minet fetch` report:
            $ minet extract -i report.csv -I downloaded > extracted.csv

        . Extracting content from a single url:
            $ minet fetch "https://lemonde.fr" | minet extract -i -

        . Indicating a custom path column (e.g. "file"):
            $ minet extract file -i report.csv -I downloaded > extracted.csv

        . Extracting content from a CSV column containing HTML directly:
            $ minet extract -i report.csv --body-column html > extracted.csv

        . Extracting content from a bunch of files using a glob pattern:
            $ minet extract "./content/**/*.html" --glob > extracted.csv

        . Extracting content using a CSV file containing glob patterns:
            $ minet extract pattern -i patterns.csv --glob > extracted.csv

        . Working on a fetch report from stdin (mind the `-`):
            $ minet fetch url file.csv | minet extract -i - -I downloaded > extracted.csv
    """,
    resolve=resolve_arguments,
    variadic_input={"dummy_column": "path", "optional": True, "no_help": True},
    resumer=IndexedResumer,
    arguments=[
        {
            "flags": ["-g", "--glob"],
            "help": "Will interpret given paths as glob patterns to resolve if given.",
            "action": "store_true",
        },
        {
            "flags": ["-I", "--input-dir"],
            "help": "Directory where the HTML files are stored. Will default to fetch default output directory.",
        },
        {
            "flags": ["-p", "--processes"],
            "help": "Number of processes to use. Defaults to roughly half of the available CPUs.",
            "type": int,
        },
        {
            "flags": ["--chunk-size"],
            "help": "Chunk size for multiprocessing.",
            "type": int,
            "default": 1,
        },
        {
            "flag": "--body-column",
            "help": "Name of the CSV column containing html bodies.",
            "default": "body",
        },
        {
            "flag": "--error-column",
            "help": "Name of the CSV column containing a fetch error.",
            "default": "fetch_error",
        },
        {
            "flag": "--status-column",
            "help": "Name of the CSV column containing HTTP status.",
            "default": "http_status",
        },
        {
            "flag": "--encoding-column",
            "help": "Name of the CSV column containing file encoding.",
            "default": "encoding",
        },
        {
            "flag": "--mimetype-column",
            "help": "Name of the CSV column containing file mimetype.",
            "default": "mimetype",
        },
        {
            "flag": "--encoding",
            "help": "Name of the default encoding to use. If not given the command will infer it for you.",
        },
        {
            "flags": ["-u", "--unordered"],
            "help": "Whether to allow the result to be written in an arbitrary order dependent on the multiprocessing scheduling. Can improve performance.",
            "action": "store_true",
        },
    ],
)



from typing import Any, cast

from bs4 import SoupStrainer, BeautifulSoup

from .types import NamedScraper


class CanonicalScraper(NamedScraper):
    name = "canonical"
    fieldnames = ["canonical_url"]
    plural = False
    output_type = "scalar"
    strainer = SoupStrainer(name="link", attrs={"rel": "canonical"})

    def scrape(self, soup: BeautifulSoup, context=None) -> Any:
        link_elem = soup.select_one("link[rel=canonical][href]")

        if link_elem is None:
            return None

        url = link_elem.get("href")

        if url is None:
            return None

        url = cast(str, url).strip()

        if not url:
            return None

        return url



# =============================================================================
# Minet Encodings
# =============================================================================
#
# List of python-supported encodings so we can ensure we will be able to
# correctly handle them.
#
from typing import Optional

import re
import charset_normalizer

ENCODINGS = set(
    [
        "1125",
        "273",
        "437",
        "646",
        "850",
        "852",
        "855",
        "857",
        "858",
        "860",
        "861",
        "862",
        "863",
        "865",
        "866",
        "869",
        "8859",
        "932",
        "936",
        "949",
        "950",
        "arabic",
        "ascii",
        "big5",
        "big5hkscs",
        "big5tw",
        "chinese",
        "cp037",
        "cp1006",
        "cp1026",
        "cp1125",
        "cp1140",
        "cp1250",
        "cp1251",
        "cp1252",
        "cp1253",
        "cp1254",
        "cp1255",
        "cp1256",
        "cp1257",
        "cp1258",
        "cp1361",
        "cp154",
        "cp273",
        "cp424",
        "cp437",
        "cp500",
        "cp65001",
        "cp720",
        "cp737",
        "cp775",
        "cp819",
        "cp850",
        "cp852",
        "cp855",
        "cp856",
        "cp857",
        "cp858",
        "cp860",
        "cp861",
        "cp862",
        "cp863",
        "cp864",
        "cp865",
        "cp866",
        "cp866u",
        "cp869",
        "cp874",
        "cp875",
        "cp932",
        "cp936",
        "cp949",
        "cp950",
        "cpgr",
        "cpis",
        "csbig5",
        "csibm273",
        "csiso2022jp",
        "csiso2022kr",
        "csiso58gb231280",
        "csptcp154",
        "csshiftjis",
        "cyrillic",
        "cyrillicasian",
        "ebcdiccpbe",
        "ebcdiccpch",
        "ebcdiccphe",
        "euccn",
        "eucgb2312cn",
        "eucjis2004",
        "eucjisx0213",
        "eucjp",
        "euckr",
        "gb18030",
        "gb180302000",
        "gb2312",
        "gb23121980",
        "gb231280",
        "gbk",
        "greek",
        "greek8",
        "hebrew",
        "hkscs",
        "hz",
        "hzgb",
        "hzgb2312",
        "ibm037",
        "ibm039",
        "ibm1026",
        "ibm1125",
        "ibm1140",
        "ibm273",
        "ibm424",
        "ibm437",
        "ibm500",
        "ibm775",
        "ibm850",
        "ibm852",
        "ibm855",
        "ibm857",
        "ibm858",
        "ibm860",
        "ibm861",
        "ibm862",
        "ibm863",
        "ibm864",
        "ibm865",
        "ibm866",
        "ibm869",
        "iso2022jp",
        "iso2022jp1",
        "iso2022jp2",
        "iso2022jp2004",
        "iso2022jp3",
        "iso2022jpext",
        "iso2022kr",
        "iso88591",
        "iso885910",
        "iso885911",
        "iso885913",
        "iso885914",
        "iso885915",
        "iso885916",
        "iso88592",
        "iso88593",
        "iso88594",
        "iso88595",
        "iso88596",
        "iso88597",
        "iso88598",
        "iso88599",
        "isoir58",
        "jisx0213",
        "johab",
        "koi8r",
        "koi8u",
        "korean",
        "ksc5601",
        "ksc56011987",
        "ksx1001",
        "l1",
        "l10",
        "l2",
        "l3",
        "l4",
        "l5",
        "l6",
        "l7",
        "l8",
        "l9",
        "latin",
        "latin1",
        "latin10",
        "latin2",
        "latin3",
        "latin4",
        "latin5",
        "latin6",
        "latin7",
        "latin8",
        "latin9",
        "maccentraleurope",
        "maccyrillic",
        "macgreek",
        "maciceland",
        "macintosh",
        "maclatin2",
        "macroman",
        "macturkish",
        "ms1361",
        "ms932",
        "ms936",
        "ms949",
        "ms950",
        "mskanji",
        "pt154",
        "ptcp154",
        "ruscii",
        "shiftjis",
        "shiftjis2004",
        "shiftjisx0213",
        "sjis",
        "sjis2004",
        "sjisx0213",
        "thai",
        "u16",
        "u32",
        "u7",
        "u8",
        "uhc",
        "ujis",
        "unicode11utf7",
        "usascii",
        "utf",
        "utf16",
        "utf16be",
        "utf16le",
        "utf32",
        "utf32be",
        "utf32le",
        "utf7",
        "utf8",
        "utf8sig",
        "windows1250",
        "windows1251",
        "windows1252",
        "windows1253",
        "windows1254",
        "windows1255",
        "windows1256",
        "windows1257",
        "windows1258",
    ]
)

CLEANER_RE = re.compile(r"[\s\-_]")


def normalize_encoding(encoding: str) -> str:
    return re.sub(CLEANER_RE, "", encoding).lower()


def is_supported_encoding(encoding):
    return normalize_encoding(encoding) in ENCODINGS


def encoding_sort_key(encoding):
    return (normalize_encoding(encoding), -len(encoding))


def infer_charset(
    data: bytes, only_supported_encodings: bool = True
) -> Optional[charset_normalizer.CharsetMatch]:
    matches = charset_normalizer.from_bytes(data)

    for match in matches:
        if not only_supported_encodings or is_supported_encoding(match.encoding):
            return match

    return None


def infer_encoding(data: bytes, only_supported_encodings: bool = True) -> Optional[str]:
    charset = infer_charset(data, only_supported_encodings=only_supported_encodings)

    if charset is None:
        return None

    return charset.encoding


UTF8_BOM = b"\xef\xbb\xbf"
UTF16_LE_BOM = b"\xff\xfe"
UTF16_BE_BOM = b"\xfe\xff"
UTF32_LE_BOM = b"\xff\xfe\x00\x00"
UTF32_BE_BOM = b"\x00\x00\xfe\xff"


def fix_surrogates(string: str) -> str:
    try:
        return string.encode("utf-16", errors="surrogatepass").decode("utf-16")
    except (UnicodeEncodeError, UnicodeDecodeError):
        return string.encode("utf-8", errors="replace").decode("utf-8")



# =============================================================================
# Minet Crawler
# =============================================================================
#
# Crawler class invoking multiple spiders to scrape data from the web.
#
from typing import (
    cast,
    overload,
    Optional,
    TypeVar,
    Callable,
    Dict,
    Tuple,
    Any,
    Generic,
    Mapping,
    Iterable,
    Iterator,
    Union,
    Awaitable,
    TYPE_CHECKING,
)
from minet.types import ParamSpec

if TYPE_CHECKING:
    from playwright.async_api import BrowserContext
    from minet.browser import ThreadsafeBrowser
    from minet.web import Response

from os import makedirs
from os.path import join
from threading import Lock
from urllib.parse import urljoin
from ural import ensure_protocol, get_domain_name
from multiprocessing import Pool

from minet.crawl.types import (
    CrawlJob,
    CrawlTarget,
    UrlOrCrawlTarget,
    CrawlJobDataType,
    CrawlResultDataType,
    SuccessfulCrawlResult,
    ErroredCrawlResult,
    AnyCrawlResult,
)
from minet.crawl.spiders import (
    Spider,
    FunctionSpider,
    FunctionSpiderCallable,
)
from minet.crawl.exceptions import (
    CrawlerAlreadyFinishedError,
    CrawlerSpiderProcessError,
)
from minet.crawl.queue import CrawlerQueue, AnyParallelism, AnyThrottle
from minet.crawl.state import CrawlerState
from minet.crawl.url_cache import URLCache
from minet.web import request, EXPECTED_WEB_ERRORS, AnyTimeout
from minet.fs import ThreadSafeFileWriter
from minet.executors import HTTPThreadPoolExecutor, CallbackResultType
from minet.exceptions import UnknownSpiderError, CancelledRequestError
from minet.constants import (
    DEFAULT_DOMAIN_PARALLELISM,
    DEFAULT_THROTTLE,
    DEFAULT_FETCH_MAX_REDIRECTS,
    DEFAULT_URLLIB3_TIMEOUT,
)

P = ParamSpec("P")
T = TypeVar("T")

DEFAULT_SPIDER_KEY = object()


def coerce_spider(target):
    if isinstance(target, Spider):
        return target

    if callable(target):
        return FunctionSpider(target)

    raise TypeError("expecting a spider or a callable")


def spider_start_iter(spider: Spider) -> Iterable[UrlOrCrawlTarget]:
    if spider.START_URL is not None:
        yield spider.START_URL

    if spider.START_URLS is not None:
        yield from spider.START_URLS

    if spider.START_TARGET is not None:
        yield spider.START_TARGET

    if spider.START_TARGETS is not None:
        yield from spider.START_TARGETS

    start = spider.start()

    if start is not None:
        yield from start


RequestArgsType = Callable[[CrawlJob[CrawlJobDataType]], Dict]


class CrawlWorker(Generic[CrawlJobDataType, CrawlResultDataType, CallbackResultType]):
    def __init__(
        self,
        crawler: "Crawler",
        *,
        request_args: Optional[RequestArgsType[CrawlJobDataType]] = None,
        use_pycurl: bool = False,
        compressed: bool = False,
        max_redirects: int = DEFAULT_FETCH_MAX_REDIRECTS,
        stateful_redirects: bool = False,
        spoof_ua: bool = False,
        known_encoding: Optional[str] = None,
        callback: Optional[
            Callable[
                [
                    "Crawler",
                    SuccessfulCrawlResult[CrawlJobDataType, CrawlResultDataType],
                ],
                CallbackResultType,
            ]
        ] = None,
    ):
        self.crawler = crawler
        self.cancel_event = self.crawler.executor.cancel_event
        self.local_context = self.crawler.executor.local_context
        self.request_args = request_args
        self.max_depth = crawler.max_depth
        self.callback = callback

        self.default_kwargs = {}

        if self.crawler.browser is None:
            self.default_kwargs = {
                "pool_manager": crawler.executor.pool_manager,
                "max_redirects": max_redirects,
                "stateful": stateful_redirects,
                "spoof_ua": spoof_ua,
                "cancel_event": crawler.executor.cancel_event,
            }

        if self.crawler.executor.retry_on_statuses is not None:
            self.default_kwargs["raise_on_statuses"] = (
                self.crawler.executor.retry_on_statuses
            )

        if use_pycurl:
            del self.default_kwargs["pool_manager"]
            self.default_kwargs["use_pycurl"] = True

        if compressed:
            self.default_kwargs["compressed"] = True

        if known_encoding is not None:
            self.default_kwargs["known_encoding"] = known_encoding

    def __call__(
        self, job: CrawlJob[CrawlJobDataType]
    ) -> Optional[
        Tuple[
            AnyCrawlResult[CrawlJobDataType, CrawlResultDataType],
            Optional[CallbackResultType],
        ]
    ]:
        # Registering work
        with self.crawler.state.task(), self.crawler.queue.group_releaser(job):
            cancel_event = self.cancel_event

            spider = self.crawler.get_spider(job.spider)

            if spider is None:
                assert job.spider is not None
                return ErroredCrawlResult(job, UnknownSpiderError(job.spider)), None

            # NOTE: crawl job must have a url and a depth at that point
            assert job.url is not None
            assert job.depth is not None

            kwargs = self.default_kwargs.copy()

            if cancel_event.is_set():
                return

            if self.request_args is not None:
                # NOTE: request_args must be threadsafe
                kwargs.update(self.request_args(job))

            if cancel_event.is_set():
                return

            request_fn = (
                request
                if self.crawler.browser is None
                else self.crawler.browser.request
            )

            # NOTE: we create an atomic unit of work that will retry both the request
            # and the subsequent spider processing
            response = None

            # NOTE: the function takes "url" so that the executor may format the warning's epilog
            def retryable_work(url: str) -> Optional[Tuple["Response", Any, Any]]:
                nonlocal response

                try:
                    response = request_fn(url, **kwargs)

                except CancelledRequestError:
                    return

                if cancel_event.is_set():
                    return

                spider_result = spider.process(job, response)

                if spider_result is not None:
                    try:
                        data, next_jobs = spider_result
                    except (ValueError, TypeError):
                        raise TypeError(
                            'Spider.process is expected to return either None or a 2-tuple containing data and next targets to enqueue. Got a "%s" instead.'
                            % spider_result.__class__.__name__
                        )
                else:
                    data = None
                    next_jobs = None

                return response, data, next_jobs

            retryer = getattr(self.local_context, "retryer", None)

            try:
                if retryer is None:
                    output = retryable_work(job.url)
                else:
                    output = retryer(retryable_work, job.url)

            except EXPECTED_WEB_ERRORS as error:
                return ErroredCrawlResult(job, error), None

            except Exception as reason:
                if response is None:
                    raise

                raise CrawlerSpiderProcessError(
                    reason=reason, job=job, response=response
                )

            # Was cancelled?
            if output is None:
                return

            response, data, next_jobs = output

            if cancel_event.is_set():
                return

            degree = 0

            if next_jobs is not None:
                if isinstance(next_jobs, (str, CrawlTarget)):
                    next_jobs = [next_jobs]

                if self.max_depth is None or job.depth < self.max_depth:
                    degree = self.crawler.enqueue(
                        next_jobs,
                        spider=spider.FORWARD_SPIDER or job.spider,
                        depth=job.depth + 1,
                        base_url=response.end_url,
                        parent=job,
                    )

            result = SuccessfulCrawlResult(job, response, data, degree)

            # NOTE: at one point we might want to retry the callback like
            # in the HTTPWorker
            callback_result = None

            if self.callback is not None:
                callback_result = self.callback(self.crawler, result)  # type: ignore

            return result, callback_result  # type: ignore


CrawlJobDataTypes = TypeVar("CrawlJobDataTypes")
CrawlResultDataTypes = TypeVar("CrawlResultDataTypes")
AnySpider = Union[
    FunctionSpiderCallable[CrawlJobDataTypes, CrawlResultDataTypes],
    Spider[CrawlJobDataTypes, CrawlResultDataTypes],
]
SpiderDeclaration = Union[
    AnySpider[CrawlJobDataTypes, CrawlResultDataTypes],
    Dict[str, AnySpider[CrawlJobDataTypes, CrawlResultDataTypes]],
]


# TODO: try creating a kwarg type for those
class Crawler(Generic[CrawlJobDataTypes, CrawlResultDataTypes]):
    executor: HTTPThreadPoolExecutor

    enqueue_lock: Lock

    browser: Optional["ThreadsafeBrowser"]

    queue: CrawlerQueue
    persistent: bool

    state: CrawlerState

    started: bool
    stopped: bool
    resuming: bool
    singular: bool

    __spiders: Dict[Union[object, str], Spider[CrawlJobDataTypes, CrawlResultDataTypes]]

    def __init__(
        self,
        spider_or_spiders: SpiderDeclaration[CrawlJobDataTypes, CrawlResultDataTypes],
        persistent_storage_path: Optional[str] = None,
        sqlar: bool = False,
        visit_urls_only_once: bool = False,
        normalized_url_cache: bool = False,
        max_depth: Optional[int] = None,
        resume: bool = False,
        lifo: bool = False,
        writer_root_directory: Optional[str] = None,
        domain_parallelism: AnyParallelism = DEFAULT_DOMAIN_PARALLELISM,
        throttle: AnyThrottle = DEFAULT_THROTTLE,
        process_pool_workers: Optional[int] = None,
        max_workers: Optional[int] = None,
        wait: bool = True,
        daemonic: bool = False,
        insecure: bool = False,
        timeout: AnyTimeout = DEFAULT_URLLIB3_TIMEOUT,
        spoof_tls_ciphers: bool = False,
        proxy: Optional[str] = None,
        retry: bool = False,
        retryer_kwargs: Optional[Dict[str, Any]] = None,
        request_args: Optional[RequestArgsType[CrawlJobDataType]] = None,
        use_pycurl: bool = False,
        compressed: bool = False,
        known_encoding: Optional[str] = None,
        max_redirects: int = DEFAULT_FETCH_MAX_REDIRECTS,
        stateful_redirects: bool = False,
        spoof_ua: bool = False,
        browser_emulation: bool = False,
        browser_kwargs: Dict[str, Any] = {},
        browser_context_init: Optional[
            Callable[["BrowserContext"], Awaitable[None]]
        ] = None,
    ):
        # Validation
        if resume and persistent_storage_path is None:
            raise TypeError("cannot resume a non-persistent crawler")

        # Browser emulation?
        self.browser = None

        if browser_emulation:
            from minet.browser import ThreadsafeBrowser

            self.browser = ThreadsafeBrowser(**browser_kwargs)

            if browser_context_init is not None:
                self.browser.run_in_default_context(browser_context_init)

        # Utilities
        self.file_writer = ThreadSafeFileWriter(writer_root_directory, sqlar=sqlar)
        self.process_pool = None

        # NOTE: if not None and not 0 basically
        if process_pool_workers:
            self.process_pool = Pool(process_pool_workers)

        # Params
        self.persistent_storage_path = persistent_storage_path
        self.persistent = persistent_storage_path is not None
        self.queue_path = None
        self.url_cache_path = None
        self.max_depth = max_depth

        if self.persistent_storage_path is not None:
            makedirs(self.persistent_storage_path, exist_ok=True)

            self.queue_path = join(self.persistent_storage_path, "queue")
            self.url_cache_path = join(self.persistent_storage_path, "urls")

        # Threading
        self.enqueue_lock = Lock()

        # Lifecycle
        self.started = False
        self.stopped = False
        self.resuming = False

        # Queue
        self.queue = CrawlerQueue(
            self.queue_path,
            resume=resume,
            group_parallelism=domain_parallelism,
            throttle=throttle,
            lifo=lifo,
        )
        self.persistent = self.queue.persistent
        self.resuming = (
            self.queue.resuming
        )  # TODO: should probably also check url cache integrity

        if self.resuming and self.queue.qsize() == 0:
            raise CrawlerAlreadyFinishedError

        # Url cache
        self.unique = visit_urls_only_once
        self.url_cache = (
            URLCache(self.url_cache_path, normalized=normalized_url_cache)
            if self.unique
            else None
        )

        # Initializing state
        self.state = CrawlerState(jobs_queued=self.queue.qsize())

        # Spiders
        if isinstance(spider_or_spiders, Mapping):
            self.__spiders = {}
            self.singular = False

            for name, spider in spider_or_spiders.items():
                self.__spiders[name] = coerce_spider(spider)
        elif isinstance(spider_or_spiders, Spider) or callable(spider_or_spiders):
            self.__spiders = {DEFAULT_SPIDER_KEY: coerce_spider(spider_or_spiders)}
            self.singular = True
        else:
            raise TypeError("expecting a single spider or a mapping of spiders")

        # Attaching spiders
        for spider in self.__spiders.values():
            spider.attach(self)

        # Own executor and imap params
        # NOTE: the process pool is initialized before the HTTPThreadPoolExecutor
        # so that we don't have potential issues related to urllib3.PoolManager
        # not being fork-safe.
        self.executor = HTTPThreadPoolExecutor(
            max_workers=max_workers,
            insecure=insecure,
            timeout=timeout,
            wait=wait,
            daemonic=daemonic,
            spoof_tls_ciphers=spoof_tls_ciphers,
            proxy=proxy,
            retry=retry,
            retryer_kwargs=retryer_kwargs,
        )

        # NOTE: buffer_size=0 is very important to avoid quenouille's optimistic
        # buffer. Remember also that this cannot work if quenouille must handle
        # group parallelism or throttle, which is why it's the crawler queue's
        # job now.
        self.imap_kwargs = {"buffer_size": 0, "panic": self.queue.unblock}

        self.worker_kwargs = {
            "request_args": request_args,
            "max_redirects": max_redirects,
            "stateful_redirects": stateful_redirects,
            "spoof_ua": spoof_ua,
            "use_pycurl": use_pycurl,
            "compressed": compressed,
            "known_encoding": known_encoding,
        }

    def __repr__(self):
        class_name = self.__class__.__name__

        return "<{class_name} {number}>".format(
            class_name=class_name, number="singular" if self.singular else "plural"
        )

    def __len__(self) -> int:
        return self.queue.qsize()

    @property
    def plural(self) -> bool:
        return not self.singular

    def get_spider(self, name: Optional[str] = None) -> Spider:
        if name is None and self.plural:
            raise TypeError("plural crawler cannot return default spider")

        if name is not None and self.singular:
            raise TypeError("singular crawler cannot return a spider by name")

        if name is None:
            name = cast(str, DEFAULT_SPIDER_KEY)

        return self.__spiders[name]

    def spiders(self) -> Iterator[Tuple[str, Spider]]:
        if self.singular:
            raise TypeError(
                "singular crawler cannot iterate over its spiders (it only has the default one, use #.get_spider to get it)"
            )

        for name, spiders in self.__spiders.items():
            assert isinstance(name, str)
            yield name, spiders

    def start(self) -> None:
        # TODO: maybe start the executor ourselves if required when quenouille moves

        if self.stopped:
            # TODO: we could but we need to mind the condition below
            raise RuntimeError("Cannot restart a crawler")

        if self.started:
            raise RuntimeError("Crawler has already started")

        # Enqueuing start jobs, only if we are not resuming
        if not self.resuming:
            # NOTE: start jobs are all buffered into memory
            # We could use a blocking queue with max size but this could prove
            # difficult to resume crawls based upon lazy iterators
            for name, spider in self.__spiders.items():
                if self.singular:
                    name = None

                assert name is None or isinstance(name, str)

                self.enqueue(spider_start_iter(spider), spider=name)

        self.started = True

    def stop(self):
        # Detaching spiders
        for spider in self.__spiders.values():
            spider.detach()

        self.stopped = True

        if self.process_pool is not None:
            self.process_pool.terminate()

        self.executor.shutdown(wait=self.executor.wait)

        self.queue.close()

        if self.url_cache:
            self.url_cache.close()

        if self.browser is not None:
            self.browser.stop()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *exc):
        self.stop()

    @overload
    def crawl(
        self,
        callback: None = ...,
    ) -> Iterator[AnyCrawlResult[CrawlJobDataTypes, CrawlResultDataTypes]]: ...

    @overload
    def crawl(
        self,
        callback: Callable[
            [
                "Crawler",
                SuccessfulCrawlResult[CrawlJobDataTypes, CrawlResultDataTypes],
            ],
            Optional[CallbackResultType],
        ],
    ) -> Iterator[
        Tuple[
            AnyCrawlResult[CrawlJobDataTypes, CrawlResultDataTypes],
            Optional[CallbackResultType],
        ]
    ]: ...

    def crawl(
        self,
        callback: Optional[
            Callable[
                [
                    "Crawler",
                    SuccessfulCrawlResult[CrawlJobDataTypes, CrawlResultDataTypes],
                ],
                Optional[CallbackResultType],
            ]
        ] = None,
    ) -> Union[
        Iterator[AnyCrawlResult[CrawlJobDataTypes, CrawlResultDataTypes]],
        Iterator[
            Tuple[
                AnyCrawlResult[CrawlJobDataTypes, CrawlResultDataTypes],
                Optional[CallbackResultType],
            ]
        ],
    ]:
        worker = CrawlWorker(self, callback=callback, **self.worker_kwargs)

        def key(job: CrawlJob) -> Optional[str]:
            return job.group

        imap_unordered = self.executor.imap_unordered(
            self.queue, worker, key=key, **self.imap_kwargs
        )

        for item in imap_unordered:
            if item is None:
                continue

            result, callback_result = item

            if callback is not None:
                yield result, callback_result  # type: ignore
            else:
                yield result

            self.queue.task_done(result.job)

        # If iterator ended properly we cleanup the queue
        self.queue.cleanup()

    def __iter__(
        self,
    ) -> Iterator[AnyCrawlResult[CrawlJobDataTypes, CrawlResultDataTypes]]:
        return self.crawl()

    def enqueue(
        self,
        target_or_targets: Union[
            UrlOrCrawlTarget[CrawlJobDataTypes],
            Iterable[UrlOrCrawlTarget[CrawlJobDataTypes]],
        ],
        spider: Optional[str] = None,
        depth: Optional[int] = None,
        base_url: Optional[str] = None,
        parent: Optional[CrawlJob[CrawlJobDataTypes]] = None,
    ) -> int:
        with self.enqueue_lock:
            if isinstance(target_or_targets, (str, CrawlTarget)):
                targets = [target_or_targets]
            else:
                targets = target_or_targets

            # NOTE: we consume jobs early and before actually enqueuing to
            # catch errors early.
            # NOTE: this means that enqueue is basically the only place allowed
            # to build CrawlJob instances and validate them.
            jobs = []

            for target in targets:
                if isinstance(target, str):
                    job = CrawlJob(url=target)
                elif isinstance(target, CrawlTarget):
                    job = CrawlJob(
                        url=target.url,
                        depth=target.depth,
                        spider=target.spider,
                        priority=target.priority,
                        data=target.data,
                    )
                else:
                    raise TypeError(
                        "attempted to enqueue a target with an invalid type %s, while expecting str or CrawlTarget"
                        % target.__class__.__name__
                    )

                job.url = ensure_protocol(job.url.strip(), "https")

                if base_url is not None:
                    job.url = urljoin(base_url, job.url)

                job.group = get_domain_name(job.url)

                if spider is not None and job.spider is None:
                    job.spider = spider

                if job.spider is not None and job.spider not in self.__spiders:
                    raise UnknownSpiderError(job.spider)

                if depth is not None:
                    job.depth = depth

                if parent is not None:
                    job.parent = parent.id

                jobs.append(job)

            # Filtering urls we already visited
            if self.url_cache is not None:
                jobs = self.url_cache.register(jobs)

            count = len(jobs)

            self.queue.put_many(jobs)
            self.state.inc_queued(count)

            return count

    def write(
        self,
        filename: str,
        contents: Union[str, bytes],
        compress: bool = False,
        relative: bool = False,
    ) -> str:
        return self.file_writer.write(
            filename, contents, compress=compress, relative=relative
        )

    def submit(self, fn: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> T:
        # NOTE: this might be a footgun!
        if self.process_pool is None:
            return fn(*args, **kwargs)

        return self.process_pool.apply(fn, args, kwargs)



from typing import Optional, List

from dataclasses import dataclass
from casanova import TabularRecord

from minet.mediacloud.utils import explode_tags


@dataclass
class MediacloudTopicStory(TabularRecord):
    guid: str
    stories_id: str
    title: str
    url: str
    language: str
    media_id: str
    media_name: str
    collect_date: str
    publish_date: str
    date_is_reliable: bool
    facebook_share_count: int
    full_text_rss: bool
    inlink_count: int
    outlink_count: int
    media_inlink_count: int
    post_count: int
    snapshots_id: str
    timespans_id: str
    next_link_id: Optional[str]

    @classmethod
    def from_payload(cls, payload, next_link_id=None) -> "MediacloudTopicStory":
        return cls(
            payload["guid"],
            payload["stories_id"],
            payload["title"],
            payload["url"],
            payload["language"],
            payload["media_id"],
            payload["media_name"],
            payload["collect_date"],
            payload["publish_date"],
            payload["date_is_reliable"],
            payload["facebook_share_count"],
            bool(payload["full_text_rss"]),
            payload["inlink_count"],
            payload["outlink_count"],
            payload["media_inlink_count"],
            payload["post_count"],
            payload["snapshots_id"],
            payload["timespans_id"],
            next_link_id,
        )


@dataclass
class MediacloudStory(TabularRecord):
    guid: str
    stories_id: str
    processed_stories_id: str
    title: str
    url: str
    language: str
    collect_date: str
    publish_date: str
    media_id: str
    media_name: str
    media_url: str
    tags: List[str]
    tag_sets: List[str]
    tags_ids: List[str]
    tag_sets_ids: List[str]

    @classmethod
    def from_payload(cls, payload) -> "MediacloudStory":
        tags = payload.get("payload_tags")

        return cls(
            payload["guid"],
            payload["stories_id"],
            payload["processed_stories_id"],
            payload["title"],
            payload["url"],
            payload["language"],
            payload["collect_date"],
            payload["publish_date"],
            payload["media_id"],
            payload["media_name"],
            payload["media_url"],
            *explode_tags(tags),
        )


@dataclass
class MediacloudMedia(TabularRecord):
    media_id: str
    media_name: str
    media_url: str
    is_healthy: bool
    is_monitored: bool
    public_notes: str
    num_stories_90: int
    num_sentences_90: int
    start_date: str
    tags: List[str]
    tag_sets: List[str]
    tags_ids: List[str]
    tag_sets_ids: List[str]

    @classmethod
    def from_payload(cls, payload) -> "MediacloudMedia":
        tags = payload.get("media_source_tags")

        return cls(
            payload["media_id"],
            payload["name"],
            payload["url"],
            payload["is_healthy"],
            payload["is_monitored"],
            payload.get("public_notes"),
            payload["num_stories_90"],
            payload["num_sentences_90"],
            payload["start_date"],
            *explode_tags(tags),
        )


@dataclass
class MediacloudFeed(TabularRecord):
    name: str
    url: str
    feeds_id: str
    type: str
    media_id: str
    active: bool
    last_attempted_download_time: float
    last_new_story_time: float
    last_successful_download_time: float

    @classmethod
    def from_payload(cls, payload) -> "MediacloudFeed":
        return cls(
            payload["name"],
            payload["url"],
            payload["payloads_id"],
            payload["type"],
            payload["media_id"],
            payload["active"],
            payload.get("last_attempted_download_time"),
            payload.get("last_new_story_time"),
            payload.get("last_successful_download_time"),
        )



# =============================================================================
# Minet Loading Bars
# =============================================================================
#
# Various loading bar utilities used by minet CLI.
#
from typing import Optional, Iterable, TypeVar, Iterator
from minet.types import TypedDict, NotRequired

from contextlib import contextmanager
from collections import OrderedDict, namedtuple
from rich.live import Live
from rich.text import Text
from rich.table import Table, Column
from rich.progress import (
    Progress,
    ProgressColumn,
    TextColumn,
    BarColumn,
    SpinnerColumn,
    Task,
)
from about_time import HumanDuration, HumanThroughput
from ebbe import format_int

from minet.utils import message_flatmap
from minet.cli.console import console

T = TypeVar("T")


class CautiousBarColumn(BarColumn):
    def render(self, task: Task):
        if task.total is None:
            self.bar_width = 15
        else:
            self.bar_width = min(40, 10 + task.total)

        if task.total is not None and task.completed > task.total:
            self.finished_style = "warning"

        return super().render(task)


class TimeElapsedColumn(ProgressColumn):
    def render(self, task: Task) -> Text:
        elapsed = task.finished_time if task.finished else task.elapsed

        if not elapsed:
            return Text("")

        return Text(
            "in " + HumanDuration(elapsed).as_human(2), style="progress.elapsed"
        )


class CompletionColumn(ProgressColumn):
    def render(self, task: Task) -> Text:
        total = format_int(task.total) if task.total is not None else "?"
        completed = format_int(task.completed).rjust(len(total))

        unit = task.fields.get("unit")
        unit_text = f" {unit}" if unit is not None else ""

        return Text(f"{completed}/{total}{unit_text}")


class PercentageColumn(ProgressColumn):
    def render(self, task: Task) -> Text:
        if task.total is None or task.total == 1:
            return Text("-")

        txt = "[progress.percentage][{task.percentage:>3.0f}%]".format(task=task)
        return Text.from_markup(txt)


class ThroughputColumn(ProgressColumn):
    def render(self, task: Task) -> Text:
        elapsed = task.finished_time if task.finished else task.elapsed

        if not elapsed or not task.completed:
            return Text("(?/s)")

        throughput = HumanThroughput(task.completed / elapsed, "").as_human(2)

        eta = ""

        if task.total is not None and task.time_remaining:
            remaining = HumanDuration(task.time_remaining).as_human(2)
            eta = f" eta: {remaining}"

        message = f"({throughput}{eta})"

        return Text(message)


class StatsColumn(ProgressColumn):
    def __init__(self, table_column=None, sort_key=None):
        super().__init__(table_column=table_column)
        self.sort_key = sort_key

    def render(self, task: Task) -> Text:
        stats = task.fields.get("stats")

        if stats is None:
            return Text("")

        stats_item_shown = 0

        item_parts = []

        values = stats.values()

        if self.sort_key is not None:
            values = sorted(values, key=self.sort_key)

        for item in values:
            txt = Text()

            count = item["count"]

            if count is None:
                continue

            stats_item_shown += 1

            txt.append(str(item["name"]), style=item["style"])
            txt.append(" ")
            txt.append(format_int(count))

            item_parts.append(txt)

        parts = []

        if task.description and stats_item_shown:
            parts.append(Text("- "))

        return Text.assemble(*parts, *Text(" ").join(item_parts))


class NestedTotalColumn(ProgressColumn):
    def render(self, task: Task) -> Optional[Text]:
        sub_total_sum = task.fields.get("sub_total_sum")

        if sub_total_sum is None:
            return None

        unit = task.fields.get("unit")
        unit_text = f" {unit}" if unit is not None else ""

        return Text(f"total: {format_int(sub_total_sum)}{unit_text}")


class SingleLineNestedTotalColumn(ProgressColumn):
    def __init__(
        self, parent, unit: Optional[str], table_column: Optional[Column] = None
    ) -> None:
        self.__parent = parent
        self.__unit = unit
        super().__init__(table_column=table_column)

    def render(self, _) -> Optional[Text]:
        sub_total_sum = self.__parent.sub_total_sum

        if sub_total_sum is None:
            return None

        unit = self.__unit
        unit_text = f" {unit}" if unit is not None else ""

        return Text(f"total: {format_int(sub_total_sum)}{unit_text}")


class StatsItem(TypedDict):
    name: str
    style: NotRequired[str]
    initial: NotRequired[int]


CatchAttempt = namedtuple("CatchAttempt", ["item", "index", "exception"])


def attempt_to_catch(catch, exc, item, index=None) -> bool:
    if catch is None:
        return True

    msg = None

    # Callable version
    if callable(catch):
        return catch(CatchAttempt(item, index, exc))

    # Mapping version
    msg_format = catch.get(type(exc))

    if msg_format is None:
        return True

    if msg_format is not None:
        msg = msg_format.format(item=item, index=index)

    if index is not None:
        console.logh("Line %i" % (index + 1), msg)
    else:
        console.print(msg)

    return False


class LoadingBar(object):
    def __init__(
        self,
        title: Optional[str] = None,
        unit: Optional[str] = None,
        total: Optional[int] = None,
        completed: Optional[int] = None,
        show_label: bool = False,
        label_format: str = "{task.description}",
        stats: Optional[Iterable[StatsItem]] = None,
        stats_sort_key=None,
        nested: bool = False,
        sub_title: Optional[str] = None,
        sub_title_style: str = "info",
        sub_unit: Optional[str] = None,
        transient: bool = False,
        refresh_per_second: float = 10,
        simple: bool = False,
    ):
        self.title = title
        self.sub_total_sum = 0
        self.nested = nested
        self.show_label = show_label
        self.transient = transient
        self.known_total = total is not None
        self.already_stopped = False
        self.simple = simple

        self.bar_column = None
        self.label_progress = None
        self.label_progress_task_id = None
        self.sub_progress = None
        self.sub_task_id = None
        self.stats_progress = None
        self.stats_task = None
        self.stats = OrderedDict()
        self.stats_are_shown = False

        self.table = Table.grid(expand=True)

        # Label line
        # NOTE: this was never really used
        if show_label:
            label_progress_columns = []

            if show_label:
                label_progress_columns.append(TextColumn(label_format))

            self.label_progress = Progress(*label_progress_columns)
            self.label_progress_task_id = self.label_progress.add_task(
                description="", total=None
            )

            if not simple:
                self.table.add_row(self.label_progress)

        # Main progress line
        self.bar_column = CautiousBarColumn(pulse_style="white")

        if simple:
            self.bar_column = BarColumn(pulse_style="white", bar_width=10)

        columns = [
            TextColumn("[progress.description]{task.description}"),
            self.bar_column,
        ]

        columns.append(CompletionColumn(Column(overflow="ellipsis", no_wrap=True)))

        if not nested or simple:
            columns.append(SpinnerColumn("dots", style=None, finished_text="·"))

        columns.append(PercentageColumn(Column(overflow="ellipsis", no_wrap=True)))
        columns.append(TimeElapsedColumn(Column(overflow="ellipsis", no_wrap=True)))
        columns.append(ThroughputColumn(Column(overflow="ellipsis", no_wrap=True)))

        if nested and simple:
            columns.append(SingleLineNestedTotalColumn(self, sub_unit))

        self.progress = Progress(*columns)
        self.table.add_row(self.progress)

        self.task_id = self.progress.add_task(
            description=title or "",
            total=total,
            unit=unit,
            completed=completed or 0,
        )

        # Nested progress line
        if nested:
            sub_columns = [
                SpinnerColumn("dots", style="", finished_text="·"),
                TextColumn(
                    "{task.description}",
                    table_column=Column(max_width=50),
                    style=sub_title_style,
                ),
                CompletionColumn(),
                TimeElapsedColumn(),
                ThroughputColumn(),
                NestedTotalColumn(),
            ]

            self.sub_progress = Progress(*sub_columns)
            self.sub_task_id = self.sub_progress.add_task(
                description=sub_title or "",
                total=None,
                sub_total_sum=self.sub_total_sum,
                unit=sub_unit,
            )

            if not simple:
                self.table.add_row(self.sub_progress)

        # Stats line
        if stats is not None:
            for item in stats:
                count = item.get("initial")

                self.stats[item["name"]] = {
                    "name": item["name"],
                    "style": item.get("style", ""),
                    "count": count,
                }

                if count is not None:
                    self.stats_are_shown = True

        self.stats_progress = Progress(StatsColumn(sort_key=stats_sort_key))
        self.stats_task_id = self.stats_progress.add_task("", stats=self.stats)

        if self.stats_are_shown and not simple:
            self.table.add_row(self.stats_progress)

        # Internal live instance
        self.live = Live(
            self.table,
            refresh_per_second=refresh_per_second,
            console=console,
            transient=self.transient,
        )

    def cursor_up(self) -> None:
        # NOTE: cursor 1up
        console.file.write("\x1b[1A")

    def set_style(self, style: str) -> None:
        if self.bar_column is None:
            return

        self.bar_column.complete_style = style
        self.bar_column.finished_style = style
        self.bar_column.pulse_style = style
        self.bar_column.style = style

    def start(self) -> None:
        self.live.start()
        self.live.refresh()

    def stop(self, erase=False) -> None:
        if self.already_stopped:
            return

        if erase:
            self.live.transient = True

        self.live.stop()
        self.already_stopped = True

    def fail(self) -> None:
        self.set_style("error")
        self.stop()

    def erase(self) -> None:
        self.stop(erase=True)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        erase = False

        if exc_type is not None:
            style = "error"

            if exc_type is KeyboardInterrupt:
                if not self.already_stopped:
                    self.cursor_up()

                style = "warning"

            elif exc_type is BrokenPipeError:
                style = "warning"
                erase = True

            if self.bar_column is not None:
                self.bar_column.complete_style = style
                self.bar_column.finished_style = style
                self.bar_column.pulse_style = style

                if not self.known_total:
                    self.bar_column.style = style
        else:
            assert self.bar_column is not None
            self.bar_column.pulse_style = "success"
            self.bar_column.style = "success"

        self.stop(erase=erase)

    @contextmanager
    def step(self, item=None, count=1, index=None, catch=None, sub_total=None):
        interrupted = False

        # TODO: we could add some key kwarg if needed

        try:
            if self.nested:
                self.nested_reset()

                if item is not None:
                    self.update(sub_title=item)

                self.set_sub_total(sub_total)
            else:
                if self.show_label and item is not None:
                    self.set_label(item)
            yield

        except BaseException as exc:
            interrupted = attempt_to_catch(catch, exc, item, index)

            if interrupted:
                raise
        finally:
            if not interrupted:
                self.advance(count)

    def track(self, iterable: Iterable[T]) -> Iterator[T]:
        for item in iterable:
            with self.step(item):
                yield item

    @contextmanager
    def nested_step(self, count=1):
        assert self.nested
        interrupted = False

        try:
            yield
        except BaseException:
            interrupted = True
            raise
        finally:
            if not interrupted:
                self.nested_advance(count)

    def advance(self, count=1):
        self.progress.update(self.task_id, advance=count)

    def nested_advance(self, count=1):
        assert self.sub_progress is not None and self.sub_task_id is not None

        self.sub_total_sum += count
        self.sub_progress.update(
            self.sub_task_id, advance=count, sub_total_sum=self.sub_total_sum
        )

    def nested_reset(self):
        assert self.sub_progress is not None and self.sub_task_id is not None
        self.sub_progress.reset(self.sub_task_id)

    def __refresh_stats(self):
        assert self.stats_progress is not None
        self.stats_progress.update(self.stats_task_id, stats=self.stats)

        if not self.simple and not self.stats_are_shown:
            self.stats_are_shown = True
            self.table.add_row(self.stats_progress)

    def set_title(self, title: str):
        self.title = title
        self.progress.update(self.task_id, description=title)

    def append_to_title(self, string: str):
        self.set_title((self.title or "") + string)

    def set_total(self, total: Optional[int] = None):
        self.progress.update(self.task_id, total=total)
        self.known_total = total is not None

    def set_sub_total(self, total: Optional[int] = None):
        assert self.sub_progress is not None and self.sub_task_id is not None
        self.sub_progress.update(self.sub_task_id, total=total)

    def set_label(self, label: str):
        assert self.label_progress is not None and self.label_progress_task_id
        self.label_progress.update(self.label_progress_task_id, description=label)

    # TODO: factorize
    def set_stat(self, name: str, count: Optional[int], style: Optional[str] = None):
        assert self.stats is not None

        if name not in self.stats:
            self.stats[name] = {"name": name, "count": count, "style": style or ""}
        else:
            item = self.stats[name]
            item["count"] = count

            if style is not None:
                item["style"] = style

        self.__refresh_stats()

    def inc_stat(self, name: str, count: int = 1, style: Optional[str] = None):
        assert self.stats is not None

        if name not in self.stats:
            self.stats[name] = {"name": name, "count": count, "style": style or ""}
        else:
            item = self.stats[name]
            if item["count"] is None:
                item["count"] = count
            else:
                item["count"] += count

            if style is not None:
                item["style"] = style

        self.__refresh_stats()

    def update(self, count=None, sub_title=None, sub_count=None, label=None, **fields):
        if count is not None:
            self.advance(count)

        if sub_count is not None:
            self.nested_advance(sub_count)

        if sub_title is not None:
            assert self.sub_progress is not None and self.sub_task_id is not None
            self.sub_progress.update(self.sub_task_id, description=sub_title)

        if label is not None:
            self.set_label(label)

        if fields:
            assert self.stats is not None

            for field, count in fields.items():
                self.stats[field]["count"] = count

            self.__refresh_stats()

    def print(self, *msg, **kwargs):
        console.print(message_flatmap(*msg), **kwargs)

    def error(self, *args, **kwargs):
        console.error(*args, **kwargs)

    def warning(self, *args, **kwargs):
        console.warning(*args, **kwargs)



from urllib3.exceptions import (
    ConnectTimeoutError,
    ReadTimeoutError,
    SSLError,
    NewConnectionError,
    ProtocolError,
    DecodeError,
    LocationValueError,
    LocationParseError,
    ProxyError,
)

from minet.exceptions import (
    UnknownEncodingError,
    InvalidURLError,
    MaxRedirectsError,
    InfiniteRedirectsError,
    SelfRedirectError,
    InvalidRedirectError,
    BadlyEncodedLocationHeaderError,
    TrafilaturaError,
    FilenameFormattingError,
    FinalTimeoutError,
    CouldNotInferEncodingError,
    InvalidStatusError,
    #
    PycurlHostResolutionError,
    PycurlTimeoutError,
    PycurlConnectionRefusedError,
    PycurlSSLError,
    PycurlReceiveError,
    PycurlSendError,
    #
    BrowserYetUnimplementedError,
    BrowserNameNotResolvedError,
    BrowserConnectionAbortedError,
    BrowserConnectionRefusedError,
    BrowserConnectionClosedError,
    BrowserTimeoutError,
    BrowserSSLError,
    BrowserHTTPResponseCodeFailureError,
    BrowserConnectionTimeoutError,
    BrowserContextAlreadyClosedError,
    BrowserSocketError,
)


def serialize_new_connection_error(error):
    msg = repr(error).lower()

    if "no route to host" in msg or "errno 113" in msg:
        return "no-route-to-host"

    if "name or service not known" in msg or "errno 8" in msg or "errno -2" in msg:
        return "unknown-host"

    if "connection refused" in msg:
        return "connection-refused"

    if "temporary failure in name resolution" in msg or "errno -3" in msg:
        return "name-resolution-failure"

    if "no address associated with hostname" in msg or "errno -5" in msg:
        return "no-address-associated-with-hostname"

    if "network is unreachable" in msg or "errno 101" in msg:
        return "unreachable-network"

    return "new-connection-error"


def serialize_protocol_error(error):
    msg = repr(error).lower()

    if "connection aborted" in msg:
        return "connection-aborted"

    if "connection refused" in msg:
        return "connection-refused"

    return "connection-error"


def serialize_decode_error(error):
    msg = repr(error)

    if "gzip" in msg:
        return "invalid-gzip"

    return "invalid-encoding"


def serialize_invalid_status_error(error: InvalidStatusError):
    return "invalid-status-%i" % error.status


ERROR_REPORTERS = {
    UnicodeDecodeError: "wrong-encoding",
    UnknownEncodingError: "unknown-encoding",
    FileNotFoundError: "file-not-found",
    InvalidURLError: "invalid-url",
    LocationValueError: "invalid-url",
    LocationParseError: "invalid-url",
    SSLError: "ssl",
    NewConnectionError: serialize_new_connection_error,
    ProtocolError: serialize_protocol_error,
    ConnectTimeoutError: "connect-timeout",
    ReadTimeoutError: "read-timeout",
    FinalTimeoutError: "timeout",
    MaxRedirectsError: "max-redirects",
    InfiniteRedirectsError: "infinite-redirects",
    SelfRedirectError: "self-redirect",
    InvalidRedirectError: "invalid-redirect",
    BadlyEncodedLocationHeaderError: "badly-encoded-location",
    DecodeError: serialize_decode_error,
    TrafilaturaError: "trafilatura-error",
    FilenameFormattingError: "filename-formatting-error",
    CouldNotInferEncodingError: "could-not-infer-encoding",
    InvalidStatusError: serialize_invalid_status_error,
    ProxyError: "proxy-error",
    #
    PycurlHostResolutionError: "unknown-host",
    PycurlTimeoutError: "timeout",
    PycurlConnectionRefusedError: "connection-refused",
    PycurlSSLError: "ssl",
    PycurlReceiveError: "receive-error",
    PycurlSendError: "send-error",
    #
    BrowserYetUnimplementedError: "unknown-browser-error",
    BrowserNameNotResolvedError: "unknown-host",
    BrowserConnectionAbortedError: "connection-aborted",
    BrowserConnectionRefusedError: "connection-refused",
    BrowserConnectionClosedError: "connection-closed",
    BrowserTimeoutError: "timeout",
    BrowserSSLError: "ssl",
    BrowserHTTPResponseCodeFailureError: "http-response-code-failure",
    BrowserConnectionTimeoutError: "connect-timeout",
    BrowserContextAlreadyClosedError: "context-already-closed",
    BrowserSocketError: "socket-error",
}


def serialize_error_as_slug(error: BaseException):
    reporter = ERROR_REPORTERS.get(type(error), repr)

    return reporter(error) if callable(reporter) else reporter



# =============================================================================
# Minet YouTube CLI Action
# =============================================================================
#
# Logic of the `yt` action.
#
from minet.cli.argparse import (
    command,
    ConfigAction,
    SplitterType,
    PartialISODatetimeType,
)

# TODO: this is a lazyloading issue
from minet.youtube.constants import (
    YOUTUBE_API_DEFAULT_SEARCH_ORDER,
    YOUTUBE_API_SEARCH_ORDERS,
)

KEY_ARGUMENT = {
    "flags": ["-k", "--key"],
    "help": "YouTube API Data dashboard API key. Can be used more than once.",
    "rc_key": ["youtube", "key"],
    "action": ConfigAction,
    "plural": True,
    "required": True,
}


def youtube_api_subcommand(*args, arguments=[], **kwargs):
    return command(*args, arguments=arguments + [KEY_ARGUMENT], **kwargs)


YOUTUBE_CAPTIONS_SUBCOMMAND = command(
    "captions",
    "minet.cli.youtube.captions",
    title="YouTube captions",
    description="Retrieve captions for the given YouTube videos.",
    epilog="""
        Examples:

        . Fetching captions for a list of videos:
            $ minet yt captions video_id -i videos.csv > captions.csv

        . Fetching French captions with a fallback to English:
            $ minet yt captions video_id -i videos.csv --lang fr,en > captions.csv
    """,
    variadic_input={
        "dummy_column": "video",
        "item_label": "video url or id",
        "item_label_plural": "video urls or ids",
    },
    arguments=[
        {
            "flag": "--lang",
            "help": 'Language (ISO code like "en") of captions to retrieve. You can specify several languages by preferred order separated by commas.',
            "default": ["en"],
            "type": SplitterType(),
        },
        {
            "flags": ["-c", "--collapse"],
            "help": "Use this flag to only emit one line per video, with the subtitle lines joined together in a single cell. Note that this means losing start & duration information of the subtitles.",
            "action": "store_true",
        },
    ],
)

YOUTUBE_CHANNEL_LINKS_SUBCOMMAND = command(
    "channel-links",
    "minet.cli.youtube.channel_links",
    title="YouTube channel links",
    description="""
        Scrape official links from channel YouTube pages.
    """,
    variadic_input={"dummy_column": "channel"},
)

YOUTUBE_CHANNEL_VIDEOS_SUBCOMMAND = youtube_api_subcommand(
    "channel-videos",
    "minet.cli.youtube.channel_videos",
    title="YouTube channel videos",
    description="""
        Retrieve metadata about all YouTube videos from one or many channel(s) using the API.

        Under the hood, this command extract the channel id from the given url or scrape the
        website to find it if necessary. Then the command uses the API to retrieve
        information about videos stored in the main playlist of the channel
        supposed to contain all the channel's videos.
    """,
    epilog="""
        Examples:

        . Fetching all the videos from a channel based on the channel's id or url:
            $ minet youtube channel-videos https://www.youtube.com/c/LinksOff -k my-api-key > linksoff_videos.csv
            $ minet youtube channel-videos https://www.youtube.com/channel/UCqnbDFdCpuN8CMEg0VuEBqA -k my-api-key > thenewyorktimes_videos.csv
            $ minet youtube channel-videos UCprclkVrNPls7PR-nHhf1Ow -k my-api-key > tonyheller_videos.csv

        . Fetching multiple channels' videos:
            $ minet youtube channel-videos channel_id -i channels_id.csv -k my-api-key > channels_videos.csv
            $ minet youtube channel-videos channel_url -i channels_url.csv -k my-api-key > channels_videos.csv
    """,
    variadic_input={"dummy_column": "channel"},
    arguments=[
        {
            "flag": "--start-time",
            "help": 'The oldest UTC datetime from which the videos will be retrieved (start-time is included). The date should have the format: "YYYY-MM-DDTHH:mm:ssZ" but incomplete dates will be completed for you e.g. "2002-04".',
            "type": PartialISODatetimeType(as_string=True),
            "default": None,
        },
        {
            "flag": "--end-time",
            "help": 'The newest UTC datetime from which the videos will be retrieved (end-time is excluded). Warning: videos more recent than end-time will still be retrieved from the API, but they will not be written in the output file. The date should have the format: "YYYY-MM-DDTHH:mm:ssZ" but incomplete dates will be completed for you e.g. "2002-04".',
            "type": PartialISODatetimeType(as_string=True),
            "default": None,
        },
    ],
)

YOUTUBE_CHANNELS_SUBCOMMAND = youtube_api_subcommand(
    "channels",
    "minet.cli.youtube.channels",
    title="YouTube Channels Command",
    description="""
        Retrieve metadata about YouTube channel from one or many name(s) using the API.

        Under the hood, this command extract the channel id from the given url or scrape the
        website to find it if necessary. Then the command uses the API to retrieve
        information about the channel.
    """,
    epilog="""
        Examples:

        . Fetching metadata from a channel based on the channel's id or url:
            $ minet youtube channels https://www.youtube.com/c/LinksOff -k my-api-key > linksoff_meta.csv
            $ minet youtube channels https://www.youtube.com/channel/UCqnbDFdCpuN8CMEg0VuEBqA -k my-api-key > thenewyorktimes_meta.csv
            $ minet youtube channels UCprclkVrNPls7PR-nHhf1Ow -k my-api-key > tonyheller_meta.csv

        . Fetching multiple channels' metadata:
            $ minet youtube channels channel_id -i channels_id.csv -k my-api-key > channels.csv
            $ minet youtube channels channel_url -i channels_url.csv -k my-api-key > channels.csv
    """,
    variadic_input={"dummy_column": "channel"},
)

YOUTUBE_COMMENTS_SUBCOMMAND = youtube_api_subcommand(
    "comments",
    "minet.cli.youtube.comments",
    title="YouTube comments",
    description="Retrieve metadata about YouTube comments using the API.",
    epilog="""
        Examples:

        . Fetching a video's comments:
            $ minet yt comments https://www.youtube.com/watch?v=7JTb2vf1OQQ -k my-api-key > comments.csv
    """,
    variadic_input={"dummy_column": "video"},
)

YOUTUBE_SEARCH_SUBCOMMAND = youtube_api_subcommand(
    "search",
    "minet.cli.youtube.search",
    title="YouTube search",
    description="""
        Search videos using the YouTube API.

        Note that, even if undocumented, the API will never return
        more than approx. 500 videos for a given query.
    """,
    epilog="""
        Examples:

        . Searching videos about birds:
            $ minet youtube search bird -k my-api-key > bird_videos.csv
    """,
    variadic_input={"dummy_column": "query", "item_label_plural": "queries"},
    arguments=[
        {
            "flags": ["-l", "--limit"],
            "help": "Maximum number of videos to retrieve per query.",
            "type": int,
        },
        {
            "flags": ["--order"],
            "help": "Order in which videos are retrieved. The default one is relevance.",
            "default": YOUTUBE_API_DEFAULT_SEARCH_ORDER,
            "choices": YOUTUBE_API_SEARCH_ORDERS,
        },
    ],
)

YOUTUBE_VIDEOS_SUBCOMMAND = youtube_api_subcommand(
    "videos",
    "minet.cli.youtube.videos",
    title="YouTube videos",
    description="Retrieve metadata about YouTube videos using the API.",
    variadic_input={"dummy_column": "video"},
)

YOUTUBE_COMMAND = command(
    "youtube",
    "minet.cli.youtube",
    "Minet YouTube Command",
    aliases=["yt"],
    description="""
        Gather data from YouTube.
    """,
    subcommands=[
        YOUTUBE_CAPTIONS_SUBCOMMAND,
        YOUTUBE_CHANNEL_LINKS_SUBCOMMAND,
        YOUTUBE_CHANNEL_VIDEOS_SUBCOMMAND,
        YOUTUBE_CHANNELS_SUBCOMMAND,
        YOUTUBE_COMMENTS_SUBCOMMAND,
        YOUTUBE_SEARCH_SUBCOMMAND,
        YOUTUBE_VIDEOS_SUBCOMMAND,
    ],
)



import casanova
from os.path import isdir

from minet.crawl.url_cache import URLCache

from minet.cli.exceptions import FatalError


def action(cli_args):
    if not isdir(cli_args.urls_dir):
        raise FatalError("queue does not exist")

    # NOTE: we are only dumping so we don't need to indicate that
    # the store is normalized or not.
    store = URLCache(cli_args.urls_dir)

    writer = casanova.writer(cli_args.output, fieldnames=["url"])

    for url in store:
        writer.writerow([url])



# =============================================================================
# Minet Twitter Constants
# =============================================================================
#
# General constants used throughout the Twitter functions.
#
from urllib3 import Timeout


TWITTER_PUBLIC_API_DEFAULT_TIMEOUT = Timeout(connect=10, read=60 * 2)
TWITTER_PUBLIC_API_AUTH_HEADER = "Bearer AAAAAAAAAAAAAAAAAAAAANRILgAAAAAAnNwIzUejRCOuH5E6I8xnZz4puTs=1Zv7ttfk8LF81IUq16cHjhLTvJu4FA33AGWWjCpTnA"

TWITTER_API_MAX_STATUSES_COUNT = 200

ADDITIONAL_TWEET_FIELDS = ["intervention_type", "intervention_text", "intervention_url"]



from typing import Optional

import time
import functools
from threading import Lock


class RateLimiter:
    """
    Naive rate limiter context manager with smooth output.

    Note that it won't work in a multi-threaded environment.

    Args:
        max_per_period (int): Maximum number of calls per period.
        period (float): Duration of a period in seconds. Defaults to 1.0.

    """

    def __init__(self, max_per_period, period=1.0, with_budget=False):
        max_per_second = max_per_period / period
        self.min_interval = 1.0 / max_per_second
        self.max_budget = period / 4
        self.budget = 0.0
        self.last_entry = None
        self.with_budget = with_budget

    def enter(self):
        self.last_entry = time.perf_counter()

    def __enter__(self):
        return self.enter()

    def exit_with_budget(self):
        assert self.last_entry is not None

        running_time = time.perf_counter() - self.last_entry

        delta = self.min_interval - running_time

        # Consuming budget
        if delta >= self.budget:
            delta -= self.budget
            self.budget = 0
        else:
            self.budget -= delta
            delta = 0

        # Do we need to sleep?
        if delta > 0:
            time.sleep(delta)
        elif delta < 0:
            self.budget -= delta

        # Clamping budget
        # TODO: this should be improved by a circular buffer of last calls
        self.budget = min(self.budget, self.max_budget)

    def exit(self):
        assert self.last_entry is not None

        running_time = time.perf_counter() - self.last_entry

        delta = self.min_interval - running_time

        if delta > 0:
            time.sleep(delta)

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self.with_budget:
            return self.exit_with_budget()

        return self.exit()


class RetryableIterator:
    """
    Iterator exposing a #.retry method that will make sure the next item
    is the same as the current one.
    """

    def __init__(self, iterator):
        self.iterator = iter(iterator)
        self.current_value = None
        self.retried = False
        self.retries = 0

    def __iter__(self):
        return self

    def __next__(self):
        if self.retried:
            self.retried = False
            return self.current_value

        self.retries = 0
        self.current_value = next(self.iterator)
        return self.current_value

    def retry(self):
        self.retries += 1
        self.retried = True


class RateLimitedIterator:
    """
    Handy iterator wrapper that will yield its items while respecting a given
    rate limit and that will not sleep needlessly when the iterator is
    finally fully consumed.
    """

    def __init__(self, iterator, max_per_period, period=1.0):
        self.iterator = RetryableIterator(iterator)
        self.rate_limiter = RateLimiter(max_per_period, period)
        self.empty = False

        try:
            self.next_value = next(self.iterator)
        except StopIteration:
            self.next_value = None
            self.empty = True

    @property
    def retries(self):
        return self.iterator.retries

    def retry(self):
        return self.iterator.retry()

    def __iter__(self):
        if self.empty:
            return

        while True:
            self.rate_limiter.enter()

            yield self.next_value

            # NOTE: if the iterator is fully consumed, this will raise StopIteration
            # and skip the exit part that could sleep needlessly
            try:
                self.next_value = next(self.iterator)
            except StopIteration:
                return

            self.rate_limiter.exit()


class RateLimiterState:
    def __init__(self, max_per_period: int, period: float = 1.0):
        max_per_second = max_per_period / period
        self.min_interval = 1.0 / max_per_second
        self.last_entry = None

    def wait_if_needed(self):
        if self.last_entry is None:
            return

        running_time = time.perf_counter() - self.last_entry
        delta = self.min_interval - running_time

        if delta > 0:
            time.sleep(delta)

    def update(self):
        self.last_entry = time.perf_counter()


class ThreadsafeBurstyRateLimiterState:
    def __init__(self, max_per_period: int, period: float = 1.0):
        self.max_per_period = max_per_period
        self.period = period

        self.current_burst = 0
        self.time_of_next_burst: Optional[float] = None

        self.lock = Lock()

    def wait_if_needed(self):
        while True:
            self.lock.acquire()

            if self.current_burst < self.max_per_period:
                if self.time_of_next_burst is None:
                    self.time_of_next_burst = time.perf_counter() + self.period

                self.current_burst += 1
                self.lock.release()
                return

            assert self.time_of_next_burst is not None

            delta = self.time_of_next_burst - time.perf_counter()

            self.time_of_next_burst = None
            self.current_burst = 0

            if delta > 0:
                time.sleep(delta)
                self.lock.release()
                continue
            else:
                self.lock.release()
                return

    # NOTE: noop
    def update(self):
        pass


def rate_limited(max_per_period, period=1.0):
    state = RateLimiterState(max_per_period, period)

    def decorate(fn):
        @functools.wraps(fn)
        def decorated(*args, **kwargs):
            state.wait_if_needed()
            result = fn(*args, **kwargs)
            state.update()

            return result

        return decorated

    return decorate


def rate_limited_from_state(state):
    def decorate(fn):
        @functools.wraps(fn)
        def decorated(*args, **kwargs):
            state.wait_if_needed()
            result = fn(*args, **kwargs)
            state.update()

            return result

        return decorated

    return decorate


def rate_limited_method(attr: str = "rate_limiter_state"):
    def decorate(fn):
        @functools.wraps(fn)
        def decorated(self, *args, **kwargs):
            state = getattr(self, attr)

            if not isinstance(
                state, (RateLimiterState, ThreadsafeBurstyRateLimiterState)
            ):
                raise ValueError

            state.wait_if_needed()
            result = fn(self, *args, **kwargs)
            state.update()

            return result

        return decorated

    return decorate



# =============================================================================
# Minet Extract Content CLI Action
# =============================================================================
#
# Logic of the extract action.
#
from typing import Optional, Dict, Iterator

from os.path import isdir
from dataclasses import dataclass
from itertools import count
from casanova import IndexedEnricher
from threading import Lock

from minet.exceptions import TrafilaturaError
from minet.extraction import extract, TrafilaturaResult
from minet.multiprocessing import LazyPool
from minet.serialization import serialize_error_as_slug
from minet.fs import read_potentially_gzipped_path
from minet.cli.utils import (
    create_fetch_like_report_iterator,
    with_enricher_and_loading_bar,
    FetchReportLikeItem,
)
from minet.cli.console import console
from minet.cli.exceptions import FatalError


@dataclass
class ExtractWorkerPayload:
    id: int
    encoding: Optional[str]
    path: Optional[str]
    text: Optional[str]


@dataclass
class ExtractResult:
    id: int
    error: Optional[Exception] = None
    data: Optional[TrafilaturaResult] = None


def worker(payload: ExtractWorkerPayload) -> ExtractResult:
    text = payload.text

    result = ExtractResult(id=payload.id)

    # Reading file
    if text is None:
        try:
            text = read_potentially_gzipped_path(
                payload.path, encoding=payload.encoding, fallback_encoding="utf-8"
            )
        except (FileNotFoundError, UnicodeDecodeError) as e:
            result.error = e
            return result

    # Attempting extraction
    try:
        result.data = extract(text)
    except TrafilaturaError as e:
        result.error = e
        return result

    return result


HEADERS = ["extract_error"] + TrafilaturaResult.fieldnames()
PADDING = [""] * len(TrafilaturaResult.fieldnames())


@with_enricher_and_loading_bar(
    headers=HEADERS,
    enricher_type="indexed",
    index_column="extract_original_index",
    title="Extracting text",
    unit="docs",
    total_from_enricher=False,
)
def action(cli_args, enricher: IndexedEnricher, loading_bar):
    if cli_args.input_dir is not None and not isdir(cli_args.input_dir):
        raise FatalError(
            'Could not find the [cyan]-I/--input-dir "{}"[/cyan] directory!'.format(
                cli_args.input_dir
            )
        )

    if not cli_args.glob:
        loading_bar.set_total(enricher.total)

    if not cli_args.has_dummy_csv and cli_args.input_dir is None:
        from minet.cli.constants import DEFAULT_CONTENT_FOLDER

        cli_args.input_dir = DEFAULT_CONTENT_FOLDER

    items = create_fetch_like_report_iterator(cli_args, enricher)

    worked_on: Dict[int, FetchReportLikeItem] = {}
    worked_on_lock = Lock()

    writer_lock = Lock()

    # Fun fact: Pool.imap consumes the given iterator in a separate thread...
    def payloads() -> Iterator[ExtractWorkerPayload]:
        current_id = count()

        for item in items:
            # Items we cannot process
            if item.error is not None:
                loading_bar.advance()
                loading_bar.inc_stat(item.error, style="error")

                with writer_lock:
                    enricher.writerow(item.index, item.row)

                continue

            item_id = next(current_id)

            with worked_on_lock:
                worked_on[item_id] = item

            yield ExtractWorkerPayload(
                id=item_id, encoding=item.encoding, path=item.path, text=item.text
            )

    pool = LazyPool(cli_args.processes)

    loading_bar.append_to_title(" (p=%i)" % pool.processes)

    warned_about_input_dir = False

    with pool:
        for result in pool.imap(
            worker,
            payloads(),
            chunksize=cli_args.chunk_size,
            unordered=cli_args.unordered,
        ):
            with loading_bar.step():
                assert isinstance(result, ExtractResult)

                with worked_on_lock:
                    item = worked_on.pop(result.id)

                row = item.row

                if result.error is not None:
                    error_code = serialize_error_as_slug(result.error)
                    loading_bar.inc_stat(error_code, style="error")

                    if not warned_about_input_dir and error_code == "file-not-found":
                        warned_about_input_dir = True

                        if cli_args.input_dir is None:
                            console.warning(
                                "did you forget to give [cyan]-I/--input-dir[/cyan]? "
                            )
                        else:
                            console.warning(
                                'are you sure [cyan]-I/--input-dir "{}"[/cyan] is the correct path?'.format(
                                    cli_args.input_dir
                                )
                            )

                    with writer_lock:
                        enricher.writerow(item.index, row, [error_code], PADDING)

                    continue

                data = result.data

                if data is None:
                    loading_bar.inc_stat("no-trafilatura-result", style="error")
                    if not cli_args.input:
                        console.warning(
                            "did you forget to use [cyan]-i/--input[/cyan]?"
                        )

                    with writer_lock:
                        enricher.writerow(
                            item.index, row, ["no-trafilatura-result"], PADDING
                        )
                else:
                    with writer_lock:
                        enricher.writerow(item.index, row, [None], data)



# =============================================================================
# Minet Facebook Constants
# =============================================================================
#
# General constants used throughout the Facebook namespace.
#
from collections import OrderedDict

from minet.rate_limiting import RateLimiterState

FACEBOOK_URL = "https://www.facebook.com"
FACEBOOK_MOBILE_URL = "https://m.facebook.com"

FACEBOOK_MOBILE_DEFAULT_THROTTLE = 2.0
FACEBOOK_WEB_DEFAULT_THROTTLE = 20.0

FACEBOOK_MOBILE_RATE_LIMITER_STATE = RateLimiterState(
    1, FACEBOOK_MOBILE_DEFAULT_THROTTLE
)
FACEBOOK_WEB_RATE_LIMITER_STATE = RateLimiterState(1, FACEBOOK_WEB_DEFAULT_THROTTLE)

FACEBOOK_REACTION_KEYS = OrderedDict(
    {
        1: "like",
        2: "love",
        3: "wow",
        4: "haha",
        7: "sad",
        8: "angry",
        11: "thankful",
        12: "pride",
        16: "care",
    }
)



from typing import Union, Callable, Awaitable

import asyncio
from tempfile import gettempdir
from os.path import expanduser, join
from playwright.async_api import (
    Error as PlaywrightError,
    TimeoutError as PlaywrightTimeoutError,
    Page,
    Response,
)

from minet.exceptions import (
    BrowserError,
    BrowserNameNotResolvedError,
    BrowserConnectionAbortedError,
    BrowserConnectionRefusedError,
    BrowserConnectionClosedError,
    BrowserTimeoutError,
    BrowserYetUnimplementedError,
    BrowserSSLError,
    BrowserHTTPResponseCodeFailureError,
    BrowserConnectionTimeoutError,
    BrowserContextAlreadyClosedError,
    BrowserSocketError,
    BrowserUnableToRetrieveContentError,
)
from minet.browser.constants import BROWSERS_PATH

AnyPlaywrightError = Union[PlaywrightError, PlaywrightTimeoutError]


def convert_playwright_error(error: AnyPlaywrightError) -> BrowserError:
    lower_message = error.message.lower()

    if isinstance(error, PlaywrightTimeoutError):
        return BrowserTimeoutError()

    if "net::ERR_NAME_NOT_RESOLVED" in error.message:
        return BrowserNameNotResolvedError()

    if "net::ERR_ABORTED" in error.message:
        return BrowserConnectionAbortedError()

    if "net::ERR_CONNECTION_REFUSED" in error.message:
        return BrowserConnectionRefusedError()

    if "net::ERR_CONNECTION_CLOSED" in error.message:
        return BrowserConnectionClosedError()

    if (
        "net::ERR_CERT_AUTHORITY_INVALID" in error.message
        or "net::ERR_CERT_COMMON_NAME_INVALID" in error.message
    ):
        return BrowserSSLError()

    if "net::ERR_HTTP_RESPONSE_CODE_FAILURE" in error.message:
        return BrowserHTTPResponseCodeFailureError()

    if "net::ERR_CONNECTION_TIMED_OUT" in error.message:
        return BrowserConnectionTimeoutError()

    if (
        "target page, context or browser has been closed" in lower_message
        or "browser closed" in lower_message
        or "maybe frame was detached?" in lower_message
    ):
        return BrowserContextAlreadyClosedError()

    if "net::ERR_SOCKET_NOT_CONNECTED" in error.message:
        return BrowserSocketError()

    if "Page.content" in error.message:
        return BrowserUnableToRetrieveContentError()

    return BrowserYetUnimplementedError()


def get_browsers_path() -> str:
    return expanduser(join("~", BROWSERS_PATH))


def get_temp_persistent_context_path() -> str:
    return join(gettempdir(), "minet-browser-temporary-persistent-context")


async def try_expect_response(
    page: Page,
    action: Callable[[], Awaitable[None]],
    response_predicate: Callable[[Response], bool],
) -> Response:
    selected_response = None
    event = asyncio.Event()

    released = False

    def release() -> None:
        nonlocal released

        if released:
            return

        released = True

        page.remove_listener("response", listener)

        if not event.is_set():
            event.set()

    def listener(response: Response) -> None:
        nonlocal selected_response

        if not response_predicate(response):
            return

        selected_response = response

        release()

    async def guarded_action() -> None:
        try:
            await action()
        except Exception:
            release()
            raise

    page.on("response", listener)

    await asyncio.gather(guarded_action(), event.wait())

    assert selected_response is not None

    return selected_response



# =============================================================================
# Minet Regex Scrapers
# =============================================================================
#
# Handful of "scrapers" able to extract information from HTML without parsing
# anything with handy regexes.
#
# Those are also typically able to work on raw bytes so one does not need to
# even decode the HTML.
#
from typing import Optional, Dict

import re
from itertools import chain
from collections import defaultdict

from minet.encodings import normalize_encoding
from minet.headers import parse_http_refresh

CHARSET_RE = re.compile(rb'<meta.*?charset=["\']*(.+?)["\'>]', flags=re.I)
PRAGMA_RE = re.compile(rb'<meta.*?content=["\']*;?charset=(.+?)["\'>]', flags=re.I)
XML_RE = re.compile(rb'^\s*<\?xml.*?encoding=["\']*(.+?)["\'>]', flags=re.I)


def extract_encodings_from_xml(
    body: bytes, chunk_size: Optional[int] = 1024
) -> Dict[str, int]:
    possibilities = defaultdict(list)

    chunk = body

    if chunk_size is not None:
        chunk = body[:chunk_size]

    matches = chain(
        re.findall(CHARSET_RE, chunk),
        re.findall(PRAGMA_RE, chunk),
        re.findall(XML_RE, chunk),
    )

    for match in matches:
        encoding = match.decode().lower()
        possibilities[normalize_encoding(encoding)].append(encoding)

    return {max(names, key=len): len(names) for names in possibilities.values()}


HREF_RE = re.compile(rb'href=(\"[^"]+|\'[^\']+|[^\s]+)>?\s?', flags=re.I)


def extract_href(value):
    m = HREF_RE.search(value)

    if not m:
        return None

    url = m.group(1)

    try:
        url = url.decode("utf-8")
    except UnicodeDecodeError:
        return None

    return url.strip("\"'") or None


CANONICAL_LINK_RE = re.compile(
    rb'<link\s*[^>]*\s+rel=(?:"\s*canonical\s*"|canonical|\'\s*canonical\s*\')\s+[^>]*\s?/?>'
)


def extract_canonical_link(html_chunk: bytes):
    m = CANONICAL_LINK_RE.search(html_chunk)

    if not m:
        return None

    return extract_href(m.group())


JAVASCRIPT_LOCATION_RE = re.compile(
    rb"""(?:window\.)?location(?:\s*=\s*|\.replace\(\s*)['"`](.*?)['"`]"""
)
ESCAPED_SLASH_RE = re.compile(rb"\\\/")


def extract_javascript_relocation(html_chunk: bytes):
    m = JAVASCRIPT_LOCATION_RE.search(html_chunk)

    if not m:
        return None

    try:
        return ESCAPED_SLASH_RE.sub(b"/", m.group(1)).decode()
    except Exception:
        return None


META_REFRESH_RE = re.compile(
    rb"""<meta\s+http-equiv=['"]?refresh['"]?\s+content=['"]?([^"']+)['">]?""",
    flags=re.I,
)


def extract_meta_refresh(html_chunk: bytes):
    m = META_REFRESH_RE.search(html_chunk)

    if not m:
        return None

    return parse_http_refresh(m.group(1))



from setuptools import setup, find_packages

with open("./README.md", "r") as f:
    long_description = f.read()

# Version
# Info: https://packaging.python.org/guides/single-sourcing-package-version/
# Example: https://github.com/pypa/warehouse/blob/64ca42e42d5613c8339b3ec5e1cb7765c6b23083/warehouse/__about__.py
meta_package = {}
with open("./minet/__version__.py") as f:
    exec(f.read(), meta_package)

packages = [pkg for pkg in find_packages() if pkg.startswith("minet")]

setup(
    name="minet",
    version=meta_package["__version__"],
    description="A webmining CLI tool & library for python.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/medialab/minet",
    author="Guillaume Plique, Pauline Breteau, Jules Farjas, Héloïse Théro, Jean Descamps, Amélie Pellé, Laura Miguel, César Pichon, Kelly Christensen",
    keywords="webmining",
    license="GPL-3.0",
    python_requires=">=3.8",
    packages=packages,
    install_requires=[
        "about-time>=4,<5",
        "beautifulsoup4>=4.7.1,<5",
        "browser-cookie3==0.19.1",
        "casanova>=2.0.1,<3",
        "charset-normalizer>=3,<4",
        "dateparser>=1.1.1",
        "ebbe>=1.13,<2",
        "json5>=0.8.5",
        "lxml == 4.9.2; platform_system == 'Darwin' and python_version <= '3.8'",
        "lxml >= 5.2.2; platform_system != 'Darwin' or python_version > '3.8'",
        "nanoid>=2,<3",
        "playwright>=1.46,<1.47",
        "playwright_stealth>=1.0.6,<2",
        "pyyaml",
        "quenouille>=1.9.1,<2",
        "rich>=13,<14",
        "rich-argparse>=1,<2",
        "soupsieve>=2.1,<3",
        "tenacity>=8,<9",
        "trafilatura>=1.12.1,<1.13",
        "twitwi>=0.19.2,<0.20",
        "ural>=1.4.0,<2",
        "urllib3>=1.26.16,<2",
    ],
    extras_require={
        ":python_version<'3.11'": ["typing_extensions>=4.3"],
    },
    entry_points={"console_scripts": ["minet=minet.cli.__main__:main"]},
    zip_safe=True,
)



# TODO: lazyloading issues
from minet.constants import COOKIE_BROWSERS

from minet.cli.argparse import command

COOKIES_COMMAND = command(
    "cookies",
    "minet.cli.cookies.cookies",
    title="Minet Cookies Command",
    description="""
        Grab cookies directly from your browsers to use them easily later
        in python scripts, for instance.
    """,
    epilog="""
        Examples:

        . Dumping cookie jar from firefox:
            $ minet cookies firefox > jar.txt

        . Dumping cookies as CSV:
            $ minet cookies firefox --csv > cookies.csv

        . Print cookie for lemonde.fr:
            $ minet cookies firefox --url https://www.lemonde.fr

        . Dump cookie morsels for lemonde.fr as CSV:
            $ minet cookies firefox --url https://www.lemonde.fr --csv > morsels.csv
    """,
    arguments=[
        {
            "name": "browser",
            "help": "Name of the browser from which to grab cookies.",
            "choices": COOKIE_BROWSERS,
        },
        {
            "flag": "--csv",
            "help": "Whether to format the output as CSV. If --url is set, will output the cookie's morsels as CSV.",
            "action": "store_true",
        },
        {
            "flag": "--url",
            "help": "If given, only returns full cookie header value for this url.",
        },
    ],
)



# =============================================================================
# Minet Youtube Search CLI Action
# =============================================================================
#
# Action searching videos using YouTube's API.
#
from itertools import islice

from minet.cli.utils import with_enricher_and_loading_bar
from minet.youtube import YouTubeAPIClient
from minet.youtube.types import YouTubeVideoSnippet


@with_enricher_and_loading_bar(
    headers=YouTubeVideoSnippet,
    title="Searching videos",
    unit="queries",
    sub_unit="videos",
    nested=True,
)
def action(cli_args, enricher, loading_bar):
    client = YouTubeAPIClient(cli_args.key)

    for row, query in enricher.cells(cli_args.column, with_rows=True):
        with loading_bar.step(query):
            searcher = client.search_videos(query, order=cli_args.order)

            if cli_args.limit:
                searcher = islice(searcher, cli_args.limit)

            for video in searcher:
                loading_bar.nested_advance()
                enricher.writerow(row, video)
